import pytest
import itertools
import asyncio

from tslumd import Tally, TallyColor, TallyType, Display


def iter_tally_colors():
    yield from TallyColor

def iter_tally_types_and_colors():
    yield from itertools.product(TallyType.all(), iter_tally_colors())



def test_tally_display_conversion(faker):
    # i = 0
    # if True:
    for _ in range(100):
        i = faker.pyint(max_value=0xfffe)
        disp = Display(index=i)
        tally = Tally(i)
        for tally_type, color in iter_tally_types_and_colors():
            brightness = faker.pyint(max_value=3)
            # print(f'{i=}, {tally_type=}, {color=}')
            disp.brightness = brightness
            tally.brightness = brightness
            assert 0 <= tally.normalized_brightness <= 1
            assert tally.normalized_brightness == brightness / 3
            assert tally_type.name is not None
            setattr(tally, tally_type.name, color)
            setattr(disp, tally_type.name, color)
            for word in faker.words(3):
                # print(f'{word=}')
                tally.text = word
                disp.text = word

                assert disp == Tally.from_display(disp) == tally
                assert disp == Display.from_tally(tally) == tally
                assert Tally.from_display(disp).normalized_brightness == tally.normalized_brightness


def test_color_merge():
    t1 = Tally(0)

    t1.set_color(TallyType.rh_tally | TallyType.lh_tally, TallyColor.RED)
    assert t1.rh_tally == TallyColor.RED
    assert t1.txt_tally == TallyColor.OFF
    assert t1.lh_tally == TallyColor.RED

    t1.set_color(TallyType.rh_tally | TallyType.lh_tally, TallyColor.GREEN)
    assert t1.rh_tally == TallyColor.GREEN
    assert t1.txt_tally == TallyColor.OFF
    assert t1.lh_tally == TallyColor.GREEN

    t1.merge_color(TallyType.all_tally, TallyColor.RED)
    assert t1.rh_tally == TallyColor.AMBER
    assert t1.txt_tally == TallyColor.RED
    assert t1.lh_tally == TallyColor.AMBER

    t1.merge_color(TallyType.all_tally, TallyColor.GREEN)
    assert t1.rh_tally == TallyColor.AMBER
    assert t1.txt_tally == TallyColor.AMBER
    assert t1.lh_tally == TallyColor.AMBER

    # Reset t1 to OFF, OFF, RED
    t1.rh_tally = TallyColor.OFF
    t1.txt_tally = TallyColor.OFF
    t1.lh_tally = TallyColor.RED

    # Another Tally with only `txt_tally` set
    t2 = Tally(1, txt_tally=TallyColor.GREEN)

    # Only `txt_tally` should change
    t1.merge(t2, TallyType.all_tally)
    assert t1.rh_tally == TallyColor.OFF
    assert t1.txt_tally == TallyColor.GREEN
    assert t1.lh_tally == TallyColor.RED

    # Reset t2 to GREEN, RED, GREEN
    # t1 is still OFF, GREEN, RED
    t2.rh_tally = TallyColor.GREEN
    t2.txt_tally = TallyColor.RED
    t2.lh_tally = TallyColor.GREEN

    t1.merge(t2, TallyType.rh_tally | TallyType.lh_tally)
    assert t1.rh_tally == TallyColor.GREEN
    assert t1.txt_tally == TallyColor.GREEN
    assert t1.lh_tally == TallyColor.AMBER

    t1.merge(t2, TallyType.all_tally)
    assert t1.rh_tally == TallyColor.GREEN
    assert t1.txt_tally == TallyColor.AMBER
    assert t1.lh_tally == TallyColor.AMBER

    t2.rh_tally = TallyColor.RED
    t1.merge(t2, TallyType.all_tally)
    assert t1.rh_tally == TallyColor.AMBER
    assert t1.txt_tally == TallyColor.AMBER
    assert t1.lh_tally == TallyColor.AMBER


def test_broadcast(faker):
    for _ in range(1000):
        i = faker.pyint(max_value=0xfffe)
        tally = Tally(i)
        assert not tally.is_broadcast
        assert not Display.from_tally(tally).is_broadcast

    tally1 = Tally(0xffff)
    tally2 = Tally.broadcast()
    assert tally1.is_broadcast
    assert tally2.is_broadcast
    assert Display.from_tally(tally1).is_broadcast
    assert Display.from_tally(tally2).is_broadcast

class Listener:
    def __init__(self):
        self.results = asyncio.Queue()
    async def get(self):
        r = await self.results.get()
        self.results.task_done()
        return r
    async def callback(self, *args, **kwargs):
        # print(f'callback: {tally=}, {props_changed=}')
        await self.results.put(tuple(args))


@pytest.mark.asyncio
async def test_update_event(faker):
    loop = asyncio.get_event_loop()
    listener = Listener()

    tally = Tally(0)
    tally.bind_async(loop, on_update=listener.callback)
    tally.text = 'foo'

    _, props_changed = await listener.get()
    assert set(props_changed) == set(['text'])

    d = dict(rh_tally=TallyColor.RED, txt_tally=TallyColor.GREEN, lh_tally=TallyColor.AMBER)
    tally.update(**d)

    _, props_changed = await listener.get()
    assert set(props_changed) == set(d.keys())


    disp = Display(index=0, text=tally.text)
    tally.update_from_display(disp)
    assert disp == tally

    _, props_changed = await listener.get()
    assert set(props_changed) == set(['rh_tally', 'txt_tally', 'lh_tally'])


    for tally_type, color in iter_tally_types_and_colors():
        attr = tally_type.name
        assert attr is not None
        should_change = getattr(tally, attr) != color
        # print(f'{tally_type=}, {color=}, {should_change=}')
        setattr(tally, attr, color)
        if should_change:
            _, props_changed = await listener.get()
            assert set(props_changed) == set([attr])
        else:
            await asyncio.sleep(.01)
            assert listener.results.empty()


@pytest.mark.asyncio
async def test_control_event(faker):
    loop = asyncio.get_event_loop()
    listener = Listener()

    disp = Display(index=0)
    tally = Tally.from_display(disp)

    tally.bind_async(loop, on_control=listener.callback)

    for _ in range(100):
        data_len = faker.pyint(min_value=1, max_value=1024)
        control_data = faker.binary(length=data_len)

        disp = Display(index=0, control=control_data)
        tally.update_from_display(disp)

        _, rx_data = await listener.get()
        assert rx_data == tally.control == disp.control == control_data
        assert disp == tally == Tally.from_display(disp)


@pytest.mark.asyncio
async def test_control_event_with_text(faker):
    loop = asyncio.get_event_loop()

    text_listener = Listener()
    ctrl_listener = Listener()

    tally_text = 'foo'

    disp = Display(index=0, text=tally_text)
    tally = Tally.from_display(disp)

    assert disp == tally

    tally.bind_async(loop,
        on_update=text_listener.callback,
        on_control=ctrl_listener.callback,
    )

    for _ in range(100):
        for word in faker.words(3):
            data_len = faker.pyint(min_value=1, max_value=1024)
            control_data = faker.binary(length=data_len)

            disp = Display(index=0, control=control_data)
            tally.update_from_display(disp)

            _, rx_data = await ctrl_listener.get()
            assert rx_data == tally.control == disp.control == control_data
            assert tally.text == tally_text

            _, props_changed = await text_listener.get()
            assert set(props_changed) == set(['control'])

            tally_text=word
            disp = Display(index=0, text=tally_text)
            tally.update_from_display(disp)

            _, props_changed = await text_listener.get()
            assert set(props_changed) == set(['text'])

            assert tally.text == tally_text
            assert tally.control == control_data



import asyncio
import pytest
import pytest_asyncio

from tslumd import (
    TallyType, TallyColor, TallyKey, Tally, Screen, UmdReceiver, UmdSender,
    Message, Display,
)

class EventListener:
    def __init__(self):
        self.results = asyncio.Queue()
    async def get(self):
        r = await asyncio.wait_for(self.results.get(), 1)
        self.results.task_done()
        return r
    def empty(self):
        return self.results.empty()
    async def callback(self, *args, **kwargs):
        await self.results.put((args, kwargs))


@pytest_asyncio.fixture
async def udp_endpoint(udp_port0):
    class Protocol(asyncio.DatagramProtocol):
        def __init__(self):
            self.queue = asyncio.Queue()
            self.connected_evt = asyncio.Event()
        def connection_made(self, transport):
            self.connected_evt.set()
        def datagram_received(self, data, addr):
            self.queue.put_nowait((data, addr))

    loop = asyncio.get_event_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: Protocol(),
        ('127.0.0.1', udp_port0),
    )
    await protocol.connected_evt.wait()
    yield (transport, protocol, udp_port0)
    transport.close()


@pytest.mark.asyncio
async def test_with_uhs_data(udp_port):

    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', udp_port)])
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    screen_index = 1

    async with receiver:
        async with sender:
            # Create initial tallies using text method
            for i in range(100):
                t_id = (screen_index, i)
                sender.set_tally_text(t_id, f'Tally-{i}')
                tx_tally = sender.tallies[t_id]
                screen = sender.screens[screen_index]
                assert screen[i] is tx_tally

                evt_args, evt_kwargs = await evt_listener.get()
                rx_tally = evt_args[0]
                assert rx_tally == tx_tally

            # Create one more tally using ``set_tally_color``
            t_id = (screen_index, 200)
            sender.set_tally_color(t_id, TallyType.lh_tally, TallyColor.GREEN)
            tx_tally = sender.tallies[t_id]
            assert screen[200] is tx_tally
            evt_args, evt_kwargs = await evt_listener.get()
            rx_tally = evt_args[0]
            assert rx_tally == tx_tally

            # Allow the sender to do a full refresh.  Nothing should have changed
            await asyncio.sleep(sender.tx_interval)
            assert evt_listener.empty()

            # Connect to ``on_tally_updated`` events
            receiver.unbind(evt_listener)
            receiver.bind_async(loop, on_tally_updated=evt_listener.callback)


            # Change each tally/tally_type color to red and check the received values
            for tx_tally in sender.tallies.values():
                for tally_type in TallyType.all():
                    sender.set_tally_color(tx_tally.id, tally_type, TallyColor.RED)

                    evt_args, evt_kwargs = await evt_listener.get()
                    rx_tally = evt_args[0]
                    assert rx_tally is receiver.tallies[tx_tally.id]
                    assert rx_tally == tx_tally

            # Change the text of the extra tally from above and check
            t_id = (screen_index, 200)
            sender.set_tally_text(t_id, 'Tally-200')
            tx_tally = sender.tallies[t_id]
            evt_args, evt_kwargs = await evt_listener.get()
            rx_tally = evt_args[0]
            assert rx_tally == tx_tally

            # Let the sender to another full refresh
            await asyncio.sleep(sender.tx_interval)
            assert evt_listener.empty()

            # Change all tally/tally_type colors, but don't wait for results yet
            for tx_tally in sender.tallies.values():
                for tally_type in TallyType.all():
                    sender.set_tally_color(tx_tally.id, tally_type, TallyColor.AMBER)
                sender.set_tally_text(tx_tally.id, f'foo-{tx_tally.index}')

            # Wait for updates from last loop to get to the receiver
            # and check the results
            _ = await evt_listener.get()
            while not evt_listener.empty():
                _ = await evt_listener.get()

            for tx_tally in sender.tallies.values():
                rx_tally = receiver.tallies[tx_tally.id]
                assert rx_tally == tx_tally


@pytest.mark.asyncio
async def test_tally_type_variations(udp_port):

    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', udp_port)])
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    tally_listener = EventListener()
    receiver.bind_async(loop, on_tally_updated=tally_listener.callback)

    tally_type_strs = ('rh', 'txt', 'lh')
    tally_types = (TallyType.rh_tally, TallyType.txt_tally, TallyType.lh_tally)


    def get_tally_colors(tally):
        d = {}
        for tt in tally_types:
            d[tt] = tally[tt]
        return d

    async def wait_for_rx(tally_type):
        tally_types = set()
        if not isinstance(tally_type, TallyType):
            tally_type = TallyType.from_str(tally_type)
        if tally_type.is_iterable:
            for tt in tally_type:
                tally_types.add(tt.name)
        else:
            tally_types.add(tally_type.name)
        props = set()
        for _ in range(len(tally_types)):
            evt_args, evt_kwargs = await tally_listener.get()
            props |= evt_args[1]
            if props == tally_types:
                break
        assert props == tally_types

    screen_index = 1

    async with receiver:
        async with sender:

            for i in range(10):
                t_id = (screen_index, i)
                expected = {key:TallyColor.OFF for key in tally_types}

                tally = None
                rx_tally = None

                for tt_str, tt in zip(tally_type_strs, tally_types):

                    sender.set_tally_color(t_id, tt_str, TallyColor.RED)
                    expected[tt] = TallyColor.RED

                    tally = sender.tallies[t_id]
                    assert get_tally_colors(tally) == expected

                    if rx_tally is None:
                        evt_args, evt_kwargs = await evt_listener.get()
                        rx_tally = evt_args[0]
                        assert rx_tally == tally
                    else:
                        await wait_for_rx(tt)
                        assert get_tally_colors(rx_tally) == expected

                    sender.set_tally_color(t_id, tt, TallyColor.GREEN)
                    expected[tt] = TallyColor.GREEN
                    assert get_tally_colors(tally) == expected

                    await wait_for_rx(tt)
                    assert get_tally_colors(rx_tally) == expected

                    sender.set_tally_color(t_id, tt_str, 'off')
                    expected[tt] = TallyColor.OFF
                    assert get_tally_colors(tally) == expected

                    await wait_for_rx(tt)
                    assert get_tally_colors(rx_tally) == expected

                    sender.set_tally_color(t_id, tt, 'red')
                    expected[tt] = TallyColor.RED
                    assert get_tally_colors(tally) == expected

                    await wait_for_rx(tt)
                    assert get_tally_colors(rx_tally) == expected

                assert tally is not None
                sender.set_tally_color(t_id, 'all', 'off')
                expected = {key:TallyColor.OFF for key in tally_types}
                assert get_tally_colors(tally) == expected
                assert tally['all'] == TallyColor.OFF

                await wait_for_rx('rh|txt|lh')
                assert get_tally_colors(rx_tally) == expected

                sender.set_tally_color(t_id, 'lh|rh', 'red')
                expected[TallyType.rh_tally] = TallyColor.RED
                expected[TallyType.lh_tally] = TallyColor.RED
                assert get_tally_colors(tally) == expected
                assert tally['all'] == TallyColor.RED

                await wait_for_rx('lh|rh')
                assert get_tally_colors(rx_tally) == expected

                sender.set_tally_color(t_id, 'txt', 'green')
                expected[TallyType.txt_tally] = TallyColor.GREEN
                assert get_tally_colors(tally) == expected

                await wait_for_rx('txt')
                assert get_tally_colors(rx_tally) == expected

                assert tally['all'] == tally['lh|txt'] == tally['rh|txt'] == TallyColor.AMBER


@pytest.mark.asyncio
async def test_broadcast_display(udp_port):
    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', udp_port)])
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    async def wait_for_receiver():
        _ = await evt_listener.get()
        while not evt_listener.empty():
            _ = await evt_listener.get()

        await asyncio.sleep(sender.tx_interval)
        assert evt_listener.empty()

    color_kw = {attr:TallyColor.RED for attr in ['rh_tally', 'txt_tally', 'lh_tally']}

    screen_index = 1

    async with receiver:
        async with sender:
            # Create initial tallies
            for i in range(10):
                t_id = (screen_index, i)
                tx_tally = sender.add_tally(t_id, **color_kw)
                screen = sender.screens[screen_index]
                assert screen[i] is tx_tally
                tx_tally.text = f'Tally-{i}'

                evt_args, evt_kwargs = await evt_listener.get()
                rx_tally = evt_args[0]
                assert rx_tally == tx_tally

            # Connect to ``on_tally_updated`` events
            receiver.unbind(evt_listener)
            receiver.bind_async(loop, on_tally_updated=evt_listener.callback)

            # Send a broadcast tally for each color setting all TallyType's to it
            for color in TallyColor.all():
                color_kw = {k:color for k in color_kw.keys()}
                await sender.send_broadcast_tally(screen_index, **color_kw)
                await wait_for_receiver()

                # Check the tally colors and make sure the text values remained
                for rx_tally in receiver.tallies.values():
                    tx_tally = sender.tallies[rx_tally.id]
                    assert rx_tally.text == tx_tally.text == f'Tally-{rx_tally.index}'
                    assert rx_tally.rh_tally == tx_tally.rh_tally == color
                    assert rx_tally.txt_tally == tx_tally.txt_tally == color
                    assert rx_tally.lh_tally == tx_tally.lh_tally == color


            # Broadcast all colors to "OFF" and set all names to 'foo'
            color_kw = {k:TallyColor.OFF for k in color_kw.keys()}
            await sender.send_broadcast_tally(screen_index, text='foo', **color_kw)
            await wait_for_receiver()

            # Check the tally colors and text values
            for rx_tally in receiver.tallies.values():
                tx_tally = sender.tallies[rx_tally.id]
                assert rx_tally.text == tx_tally.text == 'foo'
                assert rx_tally.rh_tally == tx_tally.rh_tally == TallyColor.OFF
                assert rx_tally.txt_tally == tx_tally.txt_tally == TallyColor.OFF
                assert rx_tally.lh_tally == tx_tally.lh_tally == TallyColor.OFF



            # Send broadcast tally control messages
            for control_data in [b'foo', b'bar', b'baz']:
                await sender.send_broadcast_tally(screen_index, control=control_data)
                await wait_for_receiver()

                # Check for the correct control data and ensure other values
                # remain unchanged
                for rx_tally in receiver.tallies.values():
                    tx_tally = sender.tallies[rx_tally.id]
                    assert rx_tally.control == tx_tally.control == control_data
                    assert rx_tally.text == tx_tally.text == 'foo'
                    assert rx_tally.rh_tally == tx_tally.rh_tally == TallyColor.OFF
                    assert rx_tally.txt_tally == tx_tally.txt_tally == TallyColor.OFF
                    assert rx_tally.lh_tally == tx_tally.lh_tally == TallyColor.OFF

            # Do the same as above, but using the `sender.send_broadcast_tally_control` method
            # and change one tally color
            for control_data in [b'abc', b'def', b'ghi']:
                await sender.send_broadcast_tally_control(screen_index, control_data, rh_tally=TallyColor.RED)

                await wait_for_receiver()

                # Check for the correct control data and ensure other values
                for rx_tally in receiver.tallies.values():
                    tx_tally = sender.tallies[rx_tally.id]
                    assert rx_tally.control == tx_tally.control == control_data
                    assert rx_tally.text == tx_tally.text == 'foo'
                    assert rx_tally.rh_tally == tx_tally.rh_tally == TallyColor.RED
                    assert rx_tally.txt_tally == tx_tally.txt_tally == TallyColor.OFF
                    assert rx_tally.lh_tally == tx_tally.lh_tally == TallyColor.OFF



@pytest.mark.asyncio
async def test_scontrol(faker, udp_port):
    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', udp_port)])
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_scontrol=evt_listener.callback)
    bc_listener = EventListener()
    receiver.broadcast_screen.bind_async(loop, on_control=bc_listener.callback)

    async with receiver:
        async with sender:
            for i in range(100):
                data_len = faker.pyint(min_value=1, max_value=1024)
                control_data = faker.binary(length=data_len)

                await sender.send_scontrol(screen_index=i, data=control_data)

                evt_args, evt_kwargs = await evt_listener.get()

                rx_screen, rx_data = evt_args
                assert rx_screen.index == i
                assert rx_data == control_data

                # Send broadcast
                await sender.send_broadcast_scontrol(data=control_data)

                # Wait for the broadcast screen
                evt_args, evt_kwargs = await bc_listener.get()
                rx_screen, rx_data = evt_args
                assert rx_screen is receiver.broadcast_screen
                assert rx_data == control_data

                # Wait for each currently existing screen
                num_screens = i+1
                for j in range(num_screens):
                    evt_args, evt_kwargs = await evt_listener.get()
                    rx_screen, rx_data = evt_args
                    assert rx_data == control_data


@pytest.mark.asyncio
async def test_disp_control(faker, udp_port):
    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', udp_port)])
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    add_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=add_listener.callback)

    tally_listener = EventListener()
    receiver.bind_async(loop, on_tally_control=tally_listener.callback)

    screen_index = 1

    async with receiver:
        async with sender:
            for i in range(100):
                t_id = (screen_index, i)
                sender.set_tally_text(t_id, f'Tally-{i}')
                tx_tally = sender.tallies[t_id]

                evt_args, evt_kwargs = await add_listener.get()
                rx_tally = evt_args[0]
                assert rx_tally == tx_tally


                data_len = faker.pyint(min_value=1, max_value=1024)
                control_data = faker.binary(length=data_len)

                await sender.send_tally_control(t_id, control_data)
                assert tx_tally.control == control_data

                evt_args, evt_kwargs = await tally_listener.get()
                _rx_tally, rx_data = evt_args
                assert _rx_tally is rx_tally

                assert rx_data == rx_tally.control == tx_tally.control == control_data


            t_id = (screen_index, 200)
            data_len = faker.pyint(min_value=1, max_value=1024)
            control_data = faker.binary(length=data_len)
            await sender.send_tally_control(t_id, control_data)

            tx_tally = sender.tallies[t_id]

            evt_args, evt_kwargs = await add_listener.get()
            rx_tally = evt_args[0]
            assert rx_tally.control == tx_tally.control == control_data
            assert rx_tally == tx_tally


@pytest.mark.asyncio
async def test_clients_modified_during_runtime(udp_port, udp_port0):
    loop = asyncio.get_event_loop()

    sender = UmdSender()
    receiver1 = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)
    receiver2 = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port0)

    add_listener1 = EventListener()
    add_listener2 = EventListener()
    tally_listener1 = EventListener()
    tally_listener2 = EventListener()

    async def drain_listeners():
        async def drain(listener):
            while not listener.empty():
                await listener.get()
        coros = [
            drain(add_listener1),
            drain(add_listener2),
            drain(tally_listener1),
            drain(tally_listener2),
        ]
        await asyncio.gather(*coros)

    receiver1.bind_async(loop,
        on_tally_added=add_listener1.callback,
        on_tally_updated=tally_listener1.callback,
    )
    receiver2.bind_async(loop,
        on_tally_added=add_listener2.callback,
        on_tally_updated=tally_listener2.callback,
    )

    screen_index = 1

    async with receiver1:
        async with receiver2:
            async with sender:
                # Add 10 tallies and add the first receiver to clients after the third one
                for i in range(10):
                    t_id = (screen_index, i)
                    sender.set_tally_text(t_id, f'Tally {i}')
                    if i == 2:
                        sender.clients.add(('127.0.0.1', udp_port))

                # Update the tallies and add the second receiver after the third one
                for i in range(10):
                    t_id = (screen_index, i)
                    tx_tally = sender.tallies[t_id]
                    sender.set_tally_color(t_id, 'rh|lh', 'red')
                    if i == 2:
                        sender.clients.add(('127.0.0.1', udp_port0))

                # Wait for a full update then allow the messages to come in
                await asyncio.sleep(sender.tx_interval)
                await drain_listeners()
                assert receiver1.tallies == receiver2.tallies == sender.tallies


@pytest.mark.asyncio
async def test_queued_updates_are_separate_messages(udp_endpoint, udp_port):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', endpoint_port)])

    screens = {}

    async with sender:
        # Create 10 screens with 10 tallies each
        # and trigger an update by `set_tally_text`.
        #
        # Don't await within the loop so the sender.update_queue gets filled up
        for screen_index in range(10):
            screen = Screen(screen_index)
            screens[screen_index] = screen
            for tally_index in range(10):
                t_id = (screen_index, tally_index)
                txt = f'Tally-{t_id}'
                sender.set_tally_text(t_id, txt)

        # Now give the `sender.tx_loop` a chance to process the queue
        await asyncio.sleep(.1)

        # Check each message to make sure they only have a single screen's data.
        assert not protocol.queue.empty()
        while not protocol.queue.empty():
            data, addr = await protocol.queue.get()
            protocol.queue.task_done()

            parsed, _ = Message.parse(data)
            print(f'screen {parsed.screen} disp_len={len(parsed.displays)}')
            # print(parsed)
            screen = screens[parsed.screen]
            screen.update_from_message(parsed)

        # Ensure nothing got packed incorrectly by the unique tally.id in the text field
        for screen in screens.values():
            for tally in screen:
                assert tally.text == f'Tally-{tally.id}'
                assert len(screen.tallies) == 10


@pytest.mark.asyncio
async def test_all_off_on_close(faker, udp_port):
    loop = asyncio.get_event_loop()

    sender = UmdSender(
        clients=[('127.0.0.1', udp_port)],
        all_off_on_close=True,
    )
    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    add_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=add_listener.callback)

    tally_listener = EventListener()
    receiver.bind_async(loop, on_tally_updated=tally_listener.callback)

    async with receiver:
        async with sender:
            for screen_index in range(10):
                for i in range(10):
                    t_id = (screen_index, i)
                    sender.set_tally_text(t_id, f'Tally-{i}')
                    tx_tally = sender.tallies[t_id]

                    evt_args, evt_kwargs = await add_listener.get()
                    rx_tally = evt_args[0]
                    assert rx_tally == tx_tally

                    for ttype in TallyType.all():
                        assert ttype.name is not None
                        setattr(tx_tally, ttype.name, TallyColor.RED)

                        evt_args, evt_kwargs = await tally_listener.get()
                        assert getattr(rx_tally, ttype.name) == TallyColor.RED

        # Sender is closed and should have broadcast "all-off"
        _ = await asyncio.wait_for(tally_listener.get(), timeout=1)
        while not tally_listener.empty():
            _ = await tally_listener.get()
        for rx_tally in receiver.tallies.values():
            assert rx_tally.rh_tally == TallyColor.OFF
            assert rx_tally.txt_tally == TallyColor.OFF
            assert rx_tally.lh_tally == TallyColor.OFF


@pytest.mark.asyncio
async def test_broadcast_screen_updates(udp_endpoint, udp_port):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    sender = UmdSender(clients=[('127.0.0.1', endpoint_port)])

    screens = {}
    bc_screen = Screen.broadcast()

    async with sender:

        # Create 10 screens with 10 tallies each and set their initial values to
        # `text='Tally-{tally.id}', rh_tally=TallyColor.RED`
        for screen_index in range(10):
            screen = Screen(screen_index)
            screens[screen_index] = screen
            for tally_index in range(10):
                t_id = (screen_index, tally_index)
                txt = f'Tally-{t_id}'
                sender.set_tally_text(t_id, txt)
                tx_tally = sender.tallies[t_id]

                # Wait for data from the sender and parse it manually into the screen
                data, addr = await protocol.queue.get()
                protocol.queue.task_done()
                parsed, _ = Message.parse(data)
                assert parsed.screen == screen_index
                screen.update_from_message(parsed)

                assert tally_index in screen
                tally = screen[tally_index]
                assert tally.text == txt

        # For each screen, send a broadcast tally setting `rh_tally` to `RED`
        for screen_index, screen in screens.items():
            await sender.send_broadcast_tally(screen_index, rh_tally=TallyColor.RED)

            # Wait for data and parse it again into the screen
            data, addr = await protocol.queue.get()
            protocol.queue.task_done()

            parsed, _ = Message.parse(data)
            assert parsed.screen == screen_index
            screen.update_from_message(parsed)

            for tally in screen:
                assert tally.rh_tally == TallyColor.RED

        # For each tally, send a screen-broadcast (not tally-broadcast) setting
        # `text='Broadcast-{tally.index}', rh_tally=TallyColor.GREEN`
        for tally_index in range(10):
            t_id = (0xffff, tally_index)
            txt = f'Broadcast-{tally_index}'
            tally = sender.broadcast_screen.add_tally(tally_index, text=txt)#, rh_tally=TallyColor.GREEN)
            tally.rh_tally = TallyColor.GREEN

            # Wait for data and parse it into a separate broadcast screen
            data, addr = await protocol.queue.get()
            protocol.queue.task_done()
            parsed, _ = Message.parse(data)
            assert parsed.screen == 0xffff
            bc_screen.update_from_message(parsed)

            assert tally_index in bc_screen
            bc_tally = bc_screen[tally_index]
            assert bc_tally.id == t_id
            assert bc_tally.text == txt
            assert bc_tally.rh_tally == TallyColor.GREEN

            # Parse the same screen-broadcast message into each of the 10 normal screens
            # This **should** change the tally values as well
            # (unless I'm mis-interpreting the specification)
            for screen in screens.values():
                screen.update_from_message(parsed)

                sc_tally = screen[tally_index]

                assert sc_tally.text == txt
                assert sc_tally.rh_tally == TallyColor.GREEN


        # Wait for the periodic refresh from the sender which **should** send
        # the original tally states and not the broadcast ones (?)
        #
        # That's unclear, but we definitely don't want the broadcast screen
        # sending constant updates, so let's check that doesn't happen.
        # (even though that isn't specifically stated either)
        assert protocol.queue.empty()
        await asyncio.sleep(sender.tx_interval * 2)

        assert not protocol.queue.empty()
        while not protocol.queue.empty():
            data, addr = await protocol.queue.get()
            protocol.queue.task_done()
            parsed, _ = Message.parse(data)

            assert parsed.screen in screens
            assert parsed.screen != 0xffff

            screen = screens[parsed.screen]
            screen.update_from_message(parsed)

            for tally in screen:
                assert tally.text == f'Tally-{tally.id}'
                assert tally.rh_tally == TallyColor.RED



from __future__ import annotations
try:
    from loguru import logger   # type: ignore[missing-import]
except ImportError:             # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
import asyncio
import socket
import argparse
from typing import Tuple, Iterable

from pydispatch import Dispatcher, Property, DictProperty, ListProperty

from tslumd import (
    MessageType, Message, Display, TallyColor, TallyType, TallyKey,
    Tally, Screen,
)
from tslumd.tallyobj import StrOrTallyType, StrOrTallyColor
from tslumd.utils import logger_catch

Client = Tuple[str, int] #: A network client as a tuple of ``(address, port)``

__all__ = ('Client', 'UmdSender')

class UmdProtocol(asyncio.DatagramProtocol):
    def __init__(self, sender: 'UmdSender'):
        self.sender = sender
    def connection_made(self, transport):
        logger.debug(f'transport={transport}')
        self.transport = transport
        self.sender.connected_evt.set()
    def datagram_received(self, data, addr): # pragma: no cover
        pass

class UmdSender(Dispatcher):
    """Send UMD Messages

    Messages are sent immediately when a change is made to any of the
    :class:`~.Tally` objects in :attr:`tallies`. These can be added by using
    the :meth:`add_tally` method.

    Alternatively, the :meth:`set_tally_color` and :meth:`set_tally_text` methods
    may be used.

    Arguments:
        clients: Intitial value for :attr:`clients`
        all_off_on_close: Initial value for :attr:`all_off_on_close`

    .. versionchanged:: 0.0.4
        The ``all_off_on_close`` parameter was added
    """

    screens: dict[int, Screen]
    """Mapping of :class:`~.Screen` objects by :attr:`~.Screen.index`

    .. versionadded:: 0.0.3
    """

    tallies: dict[TallyKey, Tally]
    """Mapping of :class:`~.Tally` objects by their :attr:`~.Tally.id`

    Note:
        This should not be altered directly. Use :meth:`add_tally` instead

    .. versionchanged:: 0.0.3
        The keys are now a combination of the :class:`~.Screen` and
        :class:`.Tally` indices
    """

    running: bool
    """``True`` if the client / server are running
    """

    tx_interval: float = .3
    """Interval to send tally messages, regardless of state changes
    """

    clients: set[Client]
    """Set of :data:`clients <Client>` to send messages to
    """

    all_off_on_close: bool
    """If ``True``, a broadcast message will be sent before shutdown to turn
    off all tally lights in the system. (default is ``False``)

    .. versionadded:: 0.0.4
    """

    def __init__(self,
                 clients: Iterable[Client]|None = None,
                 all_off_on_close: bool = False):
        self.clients = set()
        if clients is not None:
            for client in clients:
                self.clients.add(client)
        self.all_off_on_close = all_off_on_close
        self.screens = {}
        self.tallies = {}
        self.running = False
        self.__loop: asyncio.AbstractEventLoop|None = None
        self.__broadcast_screen: Screen|None = None
        self.__update_queue: asyncio.PriorityQueue[TallyKey|tuple[int, bool]]|None = None
        self.update_task = None
        self.tx_task: asyncio.Task|None = None
        self.__connected_evt: asyncio.Event| None = None
        self.__tx_lock: asyncio.Lock|None = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """The :class:`asyncio.BaseEventLoop` associated with the instance"""
        loop = self.__loop
        if loop is None:
            loop = self.__loop = asyncio.get_running_loop()
        return loop

    @property
    def connected_evt(self) -> asyncio.Event:
        e = self.__connected_evt
        if e is None:
            e = self.__connected_evt = asyncio.Event()
        return e

    @property
    def _tx_lock(self) -> asyncio.Lock:
        l = self.__tx_lock
        if l is None:
            l = self.__tx_lock = asyncio.Lock()
        return l

    @property
    def update_queue(self) -> asyncio.PriorityQueue[TallyKey|tuple[int, bool]]:
        q = self.__update_queue
        if q is None:
            q = self.__update_queue = asyncio.PriorityQueue()
        return q

    @property
    def broadcast_screen(self) -> Screen:
        """A :class:`~.Screen` instance created using :meth:`.Screen.broadcast`

        .. versionadded:: 0.0.3
        """
        return self._build_broadcast_screen()

    def _build_broadcast_screen(self) -> Screen:
        if self.__broadcast_screen is not None:
            return self.__broadcast_screen
        screen = self.__broadcast_screen = Screen.broadcast()
        assert screen.is_broadcast
        self.screens[screen.index] = screen
        self._bind_screen(screen)
        return screen

    async def open(self):
        """Open connections and begin data transmission
        """
        if self.running:
            return
        self._build_broadcast_screen()
        self.connected_evt.clear()
        assert self.tx_task is None
        logger.debug('UmdSender.open()')
        self.running = True
        self.transport, self.protocol = await self.loop.create_datagram_endpoint(
            lambda: UmdProtocol(self),
            family=socket.AF_INET,
        )
        for screen in self.screens.values():
            self._bind_screen(screen)
        self.tx_task = asyncio.create_task(self.tx_loop())
        logger.info('UmdSender running')

    async def close(self):
        """Stop sending to clients and close connections
        """
        if not self.running:
            return
        logger.debug('UmdSender.close()')
        self.running = False
        for screen in self.screens.values():
            screen.unbind(self)
        await self.update_queue.put((0, False))
        t = self.tx_task
        self.tx_task = None
        if t is not None:
            await t
        if self.all_off_on_close:
            logger.debug('sending all off broadcast message')
            await self.send_broadcast_tally(0xffff)
        self.transport.close()
        logger.info('UmdSender closed')

    async def send_scontrol(self, screen_index: int, data: bytes):
        """Send an :attr:`SCONTROL <.Message.scontrol>` message

        Arguments:
            screen_index: The :attr:`~.Message.screen` index for the message
            data: The data to send in the :attr:`~.Message.scontrol` field

        .. versionadded:: 0.0.2
        """
        screen = self.get_or_create_screen(screen_index)
        screen.scontrol = data

    async def send_broadcast_scontrol(self, data: bytes):
        """Send a :attr:`broadcast <.Message.is_broadcast>`
        :attr:`SCONTROL <.Message.scontrol>` message

        Arguments:
            data: The data to send in the :attr:`~.Message.scontrol` field

        .. versionadded:: 0.0.2
        """
        self.broadcast_screen.scontrol = data

    def add_tally(self, tally_id: TallyKey, **kwargs) -> Tally:
        """Create a :class:`~.Tally` object and add it to :attr:`tallies` if
        one does not exist

        If necessary, creates a :class:`~.Screen` using :meth:`get_or_create_screen`

        Arguments:
            tally_id: A tuple of (:attr:`screen_index <.Screen.index>`,
                :attr:`tally_index <.Tally.index>`)
            **kwargs: Keyword arguments passed to create the tally instance

        Raises:
            KeyError: If the given ``tally_id`` already exists

        .. versionchanged:: 0.0.3
            Chaned the ``tally_index`` parameter to ``tally_id``
        """
        if tally_id in self.tallies:
            raise KeyError(f'Tally exists for id {tally_id}')
        screen_index, tally_index = tally_id
        screen = self.get_or_create_screen(screen_index)
        tally = screen.add_tally(tally_index, **kwargs)
        return tally

    def get_or_create_tally(self, tally_id: TallyKey) -> Tally:
        """If a :class:`~.Tally` object matching the given tally id exists,
        return it. Otherwise, create it using :meth:`.Screen.get_or_create_tally`

        This method is similar to :meth:`add_tally` and it can be used to avoid
        exception handling. It does not however take keyword arguments and
        is only intended for object creation.

        .. versionadded:: 0.0.3
        """
        tally = self.tallies.get(tally_id)
        if tally is not None:
            return tally
        screen_index, tally_index = tally_id
        screen = self.get_or_create_screen(screen_index)
        tally = screen.get_or_create_tally(tally_index)
        return tally

    def get_or_create_screen(self, index_: int) -> Screen:
        """Create a :class:`~.Screen` object and add it to :attr:`screens`

        Arguments:
            index_: The screen :attr:`~.Screen.index`

        Raises:
            KeyError: If the given ``index_`` already exists

        .. versionadded:: 0.0.3
        """
        if index_ in self.screens:
            return self.screens[index_]
        screen = Screen(index_)
        self.screens[screen.index] = screen
        self._bind_screen(screen)
        return screen

    def _bind_screen(self, screen: Screen):
        screen.bind(on_tally_added=self._on_screen_tally_added)
        screen.bind_async(self.loop,
            on_tally_update=self.on_tally_updated,
            on_tally_control=self.on_tally_control,
            on_control=self.on_screen_control,
        )

    def set_tally_color(self, tally_id: TallyKey, tally_type: StrOrTallyType, color: StrOrTallyColor):
        """Set the tally color for the given index and tally type

        Uses :meth:`.Tally.set_color`. See the method documentation for details

        Arguments:
            tally_id (TallyKey): A tuple of (:attr:`screen_index <.Screen.index>`,
                :attr:`tally_index <.Tally.index>`)
            tally_type (TallyType or str): :class:`~.common.TallyType` or member
                name as described in :meth:`.Tally.set_color`
            color (TallyColor or str): :class:`~.common.TallyColor` or color
                name as described in :meth:`.Tally.set_color`

        .. versionchanged:: 0.0.3
            Chaned the ``tally_index`` parameter to ``tally_id``

        .. versionchanged:: 0.0.5
            Accept string arguments and match behavior of :meth:`.Tally.set_color`
        """
        if tally_type == TallyType.no_tally:
            raise ValueError()
        tally = self.get_or_create_tally(tally_id)
        tally[tally_type] = color

    def set_tally_text(self, tally_id: TallyKey, text: str):
        """Set the tally text for the given id

        Arguments:
            tally_id: A tuple of (:attr:`screen_index <.Screen.index>`,
                :attr:`tally_index <.Tally.index>`)
            text: The :attr:`~.Tally.text` to set

        .. versionchanged:: 0.0.3
            Chaned the ``tally_index`` parameter to ``tally_id``
        """
        tally = self.get_or_create_tally(tally_id)
        tally.text = text

    async def send_tally_control(self, tally_id: TallyKey, data: bytes):
        """Send :attr:`~.Display.control` data for the given screen and tally index

        Arguments:
            tally_id: A tuple of (:attr:`screen_index <.Screen.index>`,
                :attr:`tally_index <.Tally.index>`)
            control: The control data to send

        .. versionadded:: 0.0.2

        .. versionchanged:: 0.0.3
            Chaned the ``tally_index`` parameter to ``tally_id``
        """
        tally = self.get_or_create_tally(tally_id)
        tally.control = data

    async def send_broadcast_tally_control(self, screen_index: int, data: bytes, **kwargs):
        """Send :attr:`~.Display.control` data as
        :attr:`broadcast <.Display.is_broadcast>` to all listening displays

        Arguments:
            screen_index: The screen :attr:`~.Screen.index`
            **kwargs: Additional keyword arguments to pass to the :class:`~.Tally`
                constructor

        .. versionadded:: 0.0.2

        .. versionchanged:: 0.0.3
            Added the screen_index parameter
        """
        await self.send_broadcast_tally(screen_index, control=data, **kwargs)

    async def send_broadcast_tally(self, screen_index: int, **kwargs):
        """Send a :attr:`broadcast <.Display.is_broadcast>` update
        to all listening displays

        Arguments:
            screen_index: The screen :attr:`~.Screen.index`
            **kwargs: The keyword arguments to pass to the :class:`~.Tally`
                constructor

        .. versionadded:: 0.0.2

        .. versionchanged:: 0.0.3
            Added the screen_index parameter
        """
        screen = self.get_or_create_screen(screen_index)
        tally = screen.broadcast_tally(**kwargs)
        if tally.text == '' or tally.control != b'':
            msg_type = MessageType.control
        else:
            msg_type = MessageType.display
        msg = self._build_message(screen=screen_index)
        disp = Display.from_tally(tally, msg_type=msg_type)
        msg.displays.append(disp)
        async with self._tx_lock:
            await self.send_message(msg)
            screen.unbind(self)
            for oth_tally in screen:
                oth_tally.update_from_display(disp)
            self._bind_screen(screen)

    async def on_tally_updated(self, tally: Tally, props_changed: set[str], **kwargs):
        if self.running:
            if set(props_changed) == set(['control']):
                return
            logger.debug(f'tally update: {tally}')
            await self.update_queue.put(tally.id)

    async def on_tally_control(self, tally: Tally, data: bytes, **kwargs):
        if self.running:
            async with self._tx_lock:
                disp = Display.from_tally(tally, msg_type=MessageType.control)
                assert tally.screen is not None
                msg = self._build_message(
                    screen=tally.screen.index,
                    displays=[disp],
                )
                await self.send_message(msg)


    async def on_screen_control(self, screen: Screen, data: bytes, **kwargs):
        if self.running:
            async with self._tx_lock:
                msg = self._build_message(
                    screen=screen.index,
                    type=MessageType.control,
                    scontrol=data,
                )
                await self.send_message(msg)


    def _on_screen_tally_added(self, tally: Tally, **kwargs):
        self.tallies[tally.id] = tally
        logger.debug(f'new tally: {tally}')

    @logger_catch
    async def tx_loop(self):
        async def get_queue_item(timeout):
            try:
                item = await asyncio.wait_for(self.update_queue.get(), timeout)
                if item[1] is False:
                    return False
            except asyncio.TimeoutError:
                item = None
            return item

        await self.connected_evt.wait()

        while self.running:
            item = await get_queue_item(self.tx_interval)
            if item is False:
                self.update_queue.task_done()
                break
            elif item is None:
                if not self._tx_lock.locked():
                    await self.send_full_update()
            else:
                screen_index, _ = item
                ids = set([item])
                self.update_queue.task_done()
                while not self.update_queue.empty():
                    try:
                        item = self.update_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    if item is False:
                        self.update_queue.task_done()
                        return
                    _screen_index, _ = item
                    if _screen_index == screen_index:
                        ids.add(item)
                        self.update_queue.task_done()
                    else:
                        await self.update_queue.put(item)
                        break

                msg = self._build_message(screen=screen_index)
                tallies = {i:self.tallies[i] for i in ids}
                async with self._tx_lock:
                    for key in sorted(tallies.keys()):
                        tally = tallies[key]
                        msg.displays.append(Display.from_tally(tally))
                    await self.send_message(msg)

    async def send_message(self, msg: Message):
        assert self._tx_lock.locked()
        for data in msg.build_messages():
            for client in self.clients:
                self.transport.sendto(data, client)

    async def send_full_update(self):
        coros = set()
        for screen in self.screens.values():
            coros.add(self.send_screen_update(screen))
        if not len(coros): # pragma: no cover
            return
        async with self._tx_lock:
            await asyncio.gather(*coros)

    async def send_screen_update(self, screen: Screen):
        if screen.is_broadcast:
            return
        msg = self._build_message(screen=screen.index)
        for tally in screen:
            disp = Display.from_tally(tally)
            msg.displays.append(disp)
        await self.send_message(msg)

    def _build_message(self, **kwargs) -> Message:
        return Message(**kwargs)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


class ClientArgAction(argparse._AppendAction):
    _default_help = ' '.join([
        'Client(s) to send UMD messages to formatted as "<hostaddr>:<port>".',
        'Multiple arguments may be given.',
        'If nothing is provided, defaults to "127.0.0.1:65000"',
    ])
    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=[('127.0.0.1', 65000)],
                 type_=str,
                 choices=None,
                 required=False,
                 help=_default_help,
                 metavar=None):
        super().__init__(
            option_strings, dest, nargs, const, default,
            type_, choices, required, help, metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        addr, port = values.split(':')              # type: ignore
        values = (addr, int(port))
        items = getattr(namespace, self.dest, None)
        if items == [('127.0.0.1', 65000)]:
            items = []
        else:
            items = argparse._copy_items(items)     # type: ignore
        items.append(values)
        setattr(namespace, self.dest, items)

def main():
    p = argparse.ArgumentParser()
    p.add_argument(
        '-c', '--client', dest='clients', action=ClientArgAction#, type=str,
    )
    args = p.parse_args()

    logger.info(f'Sending to clients: {args.clients!r}')

    loop = asyncio.get_event_loop()
    sender = UmdSender(clients=args.clients)
    loop.run_until_complete(sender.open())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(sender.close())
    finally:
        loop.close()

if __name__ == '__main__':
    main()



"""Implementation of the `UMDv5.0 Protocol`_ by `TSL Products`_ for tally
and other production display/control purposes.

.. _UMDv5.0 Protocol: https://tslproducts.com/media/1959/tsl-umd-protocol.pdf
.. _TSL Products: https://tslproducts.com
"""

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    # Python <3.8, fallback
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("tslumd")
except PackageNotFoundError:
    __version__ = "unknown"


try:
    from loguru import logger
except ImportError:  # pragma: no cover
    import logging

    logging.basicConfig(
        format="%(asctime)s\t%(levelname)s\t%(message)s", level=logging.DEBUG
    )
    logger = logging.getLogger(__name__)
from .common import *
from .tallyobj import *
from .messages import *
from .receiver import *
from .sender import *



import asyncio
import pytest
import pytest_asyncio

from tslumd import (
    TallyType, TallyColor, TallyKey, Message, Display,
    Tally, Screen, UmdReceiver,
)
from tslumd.messages import ParseError

class EventListener:
    def __init__(self):
        self.results = asyncio.Queue()
    async def get(self):
        r = await self.results.get()
        self.results.task_done()
        return r
    def empty(self):
        return self.results.empty()
    async def callback(self, *args, **kwargs):
        await self.results.put((args, kwargs))


@pytest_asyncio.fixture
async def udp_endpoint(udp_port0):
    class Protocol(asyncio.DatagramProtocol):
        def __init__(self):
            self.queue = asyncio.Queue()
            self.connected_evt = asyncio.Event()
        def connection_made(self, transport):
            self.connected_evt.set()
        def datagram_received(self, data, addr):
            self.queue.put_nowait((data, addr))

    loop = asyncio.get_event_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: Protocol(),
        ('127.0.0.1', udp_port0),
    )
    await protocol.connected_evt.wait()
    yield (transport, protocol, udp_port0)
    transport.close()


@pytest.mark.asyncio
async def test_with_uhs_data(uhs500_msg_bytes, uhs500_msg_parsed, udp_endpoint, udp_port):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    uhs_screen = uhs500_msg_parsed.screen

    async with receiver:

        # Send message bytes to receiver
        transport.sendto(uhs500_msg_bytes, ('127.0.0.1', udp_port))

        # Wait for all ``on_tally_added`` events
        _ = await evt_listener.get()
        while not evt_listener.empty():
            _ = await evt_listener.get()

        screen = receiver.screens[uhs_screen]

        # Check all receiver tallies against the expected ones
        assert len(receiver.tallies) == len(uhs500_msg_parsed.displays)

        for disp in uhs500_msg_parsed.displays:
            assert disp.index in screen
            tally = screen.tallies[disp.index]
            assert tally.id == (uhs_screen, disp.index)
            assert receiver.tallies[tally.id] is tally
            assert disp == tally

        # Change each display and send the updated message to receiver
        # Then wait for ``on_tally_updated`` events
        receiver.unbind(evt_listener)
        receiver.bind_async(loop, on_tally_updated=evt_listener.callback)
        for disp in uhs500_msg_parsed.displays:
            tally = screen.tallies[disp.index]

            for tally_type in TallyType.all():
                attr = tally_type.name
                assert attr is not None
                cur_value = getattr(disp, attr)
                if cur_value == TallyColor.RED:
                    new_value = TallyColor.GREEN
                else:
                    new_value = TallyColor.RED
                setattr(disp, attr, new_value)
            disp.text = f'{disp.text}-foo'
            disp.brightness = 1

            transport.sendto(uhs500_msg_parsed.build_message(), ('127.0.0.1', udp_port))

            evt_args, evt_kwargs = await evt_listener.get()
            evt_tally = evt_args[0]
            assert evt_tally is tally
            assert disp == tally


@pytest.mark.asyncio
async def test_broadcast_display(uhs500_msg_bytes, uhs500_msg_parsed, udp_endpoint, udp_port, faker):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    displays_by_index = {disp.index: disp for disp in uhs500_msg_parsed.displays}

    async with receiver:
        # Populate the receiver's tallies and wait for them to be added
        transport.sendto(uhs500_msg_bytes, ('127.0.0.1', udp_port))
        _ = await evt_listener.get()
        while not evt_listener.empty():
            _ = await evt_listener.get()

        assert len(receiver.tallies) == len(uhs500_msg_parsed.displays)

        receiver.unbind(evt_listener)
        receiver.bind_async(loop, on_tally_updated=evt_listener.callback)


        # Send a broadcast display message for each TallyColor with a random brightness
        # The control field is set so the Tally text field is unchanged
        for color in TallyColor:
            brightness = faker.pyint(max_value=3)
            bc_disp = Display.broadcast(
                control=b'foo', rh_tally=color, lh_tally=color, txt_tally=color,
                brightness=brightness,
            )
            msg = Message(displays=[bc_disp])
            transport.sendto(msg.build_message(), ('127.0.0.1', udp_port))

            _ = await evt_listener.get()
            while not evt_listener.empty():
                _ = await evt_listener.get()

            # Check each of the receiver's tallies against the bc_disp values
            # and make sure the text didn't change
            for tally in receiver.tallies.values():
                assert tally.index < 0xffff
                assert not tally.is_broadcast
                assert tally.rh_tally == color
                assert tally.txt_tally == color
                assert tally.lh_tally == color
                assert tally.control == bc_disp.control
                assert tally.brightness == brightness
                assert tally.text == displays_by_index[tally.index].text


@pytest.mark.asyncio
async def test_rebind(
    uhs500_msg_bytes,
    uhs500_msg_parsed,
    udp_endpoint,
    udp_port,
    unused_udp_port_factory,
    non_loopback_hostaddr
):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    host_addrs = ['127.0.0.1', non_loopback_hostaddr]

    loop = asyncio.get_event_loop()

    receiver = UmdReceiver(hostaddr=host_addrs[0], hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    async with receiver:

        # Send message bytes to receiver and wait for ``on_tally_added`` events
        transport.sendto(uhs500_msg_bytes, (host_addrs[0], udp_port))
        _ = await evt_listener.get()
        while not evt_listener.empty():
            _ = await evt_listener.get()

        assert len(receiver.tallies) == len(uhs500_msg_parsed.displays)

        # Set up to get ``on_tally_updated`` callbacks
        receiver.unbind(evt_listener)
        receiver.bind_async(loop, on_tally_updated=evt_listener.callback)


        # Change bind address and trigger a change
        await receiver.set_hostaddr(host_addrs[1])
        assert receiver.hostaddr == host_addrs[1]

        disp = uhs500_msg_parsed.displays[0]
        disp.brightness = 1

        transport.sendto(uhs500_msg_parsed.build_message(), (host_addrs[1], udp_port))

        evt_args, evt_kwargs = await evt_listener.get()
        evt_tally = evt_args[0]
        assert disp == evt_tally

        # Change bind port and trigger a change
        new_port = unused_udp_port_factory()
        assert new_port != udp_port

        await receiver.set_hostport(new_port)
        assert receiver.hostport == new_port

        disp.brightness = 2
        transport.sendto(uhs500_msg_parsed.build_message(), (host_addrs[1], new_port))

        evt_args, evt_kwargs = await evt_listener.get()
        evt_tally = evt_args[0]
        assert disp == evt_tally


@pytest.mark.asyncio
async def test_scontrol(faker, udp_endpoint, udp_port):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_scontrol=evt_listener.callback)

    async with receiver:
        for i in range(100):
            data_len = faker.pyint(min_value=1, max_value=1024)
            control_data = faker.binary(length=data_len)

            msgobj = Message(screen=i, scontrol=control_data)
            transport.sendto(msgobj.build_message(), ('127.0.0.1', udp_port))

            args, kwargs = await evt_listener.get()

            screen, rx_data = args
            assert screen.index == i
            assert rx_data == control_data


@pytest.mark.asyncio
async def test_parse_errors(faker, udp_endpoint, udp_port):
    transport, protocol, endpoint_port = udp_endpoint
    assert udp_port != endpoint_port

    loop = asyncio.get_event_loop()

    receiver = UmdReceiver(hostaddr='127.0.0.1', hostport=udp_port)

    evt_listener = EventListener()
    receiver.bind_async(loop, on_tally_added=evt_listener.callback)

    msgobj = Message()
    disp = Display(index=0, text='foo')
    msgobj.displays.append(disp)

    async with receiver:
        transport.sendto(msgobj.build_message(), ('127.0.0.1', udp_port))

        _ = await evt_listener.get()

        receiver.unbind(evt_listener)
        receiver.bind_async(loop, on_tally_updated=evt_listener.callback)

        screen = receiver.screens[msgobj.screen]
        rx_disp = screen[disp.index]
        assert rx_disp is receiver.tallies[rx_disp.id]
        assert rx_disp == disp

        for i in range(100):
            num_bytes = faker.pyint(min_value=1, max_value=1024)
            bad_bytes = faker.binary(length=num_bytes)

            with pytest.raises(ParseError):
                receiver.parse_incoming(bad_bytes, ('127.0.0.1', endpoint_port))

            transport.sendto(bad_bytes, ('127.0.0.1', udp_port))

            disp.text = f'foo_{i}'
            transport.sendto(msgobj.build_message(), ('127.0.0.1', udp_port))

            _ = await evt_listener.get()

            assert rx_disp == disp



from __future__ import annotations
try:
    from loguru import logger   # type: ignore[missing-import]
except ImportError:             # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
import asyncio
from typing import Tuple

from pydispatch import Dispatcher, Property, DictProperty, ListProperty

from tslumd import Tally, Screen, TallyKey, Message

__all__ = ('UmdReceiver',)




class UmdProtocol(asyncio.DatagramProtocol):
    def __init__(self, receiver: 'UmdReceiver'):
        self.receiver = receiver
    def connection_made(self, transport):
        logger.debug(f'transport={transport}')
        self.transport = transport
        self.receiver.connected_evt.set()
    def datagram_received(self, data, addr):
        self.receiver.parse_incoming(data, addr)


class UmdReceiver(Dispatcher):
    """Receiver for UMD messages

    Arguments:
        hostaddr (str): The local host address to bind the server to. Defaults
            to :attr:`DEFAULT_HOST`
        hostport (int): The port to listen on. Defaults to :attr:`DEFAULT_PORT`

    :Events:
        .. event:: on_tally_added(tally: Tally)

            Fired when a :class:`~.Tally` instance is added to :attr:`tallies`

        .. event:: on_tally_updated(tally: Tally)

            Fired when any :class:`~.Tally` instance has been updated

        .. event:: on_tally_control(tally: Tally, data: bytes)

            Fired when control data has been received for a :class:`~.Tally`

            .. versionadded:: 0.0.3

        .. event:: on_screen_added(screen: Screen)

            Fired when a :class:`~.Screen` instance is added to :attr:`screens`

            .. versionadded:: 0.0.3

        .. event:: on_scontrol(screen: int, data: bytes)

            Fired when a message with :attr:`~.Message.scontrol` data is received

            * ``screen`` : The :attr:`~.Message.screen` from the incoming
              control message
            * ``data`` : The control data

            .. versionadded:: 0.0.2

    """

    DEFAULT_HOST: str = '0.0.0.0' #: The default host address to listen on
    DEFAULT_PORT: int = 65000 #: The default host port to listen on

    screens: dict[int, Screen]
    """Mapping of :class:`~.Screen` objects by :attr:`~.Screen.index`

    .. versionadded:: 0.0.3
    """

    broadcast_screen: Screen
    """A :class:`~.Screen` instance created using :meth:`.Screen.broadcast`

    .. versionadded:: 0.0.3
    """

    tallies: dict[TallyKey, Tally]
    """Mapping of :class:`~.Tally` objects by their :attr:`~.Tally.id`

    .. versionchanged:: 0.0.3
        The keys are now a combination of the :class:`~.Screen` and
        :class:`.Tally` indices
    """

    running: bool
    """``True`` if the client / server are running
    """

    _events_ = [
        'on_tally_added', 'on_tally_updated', 'on_tally_control',
        'on_screen_added', 'on_scontrol',
    ]
    def __init__(self, hostaddr: str = DEFAULT_HOST, hostport: int = DEFAULT_PORT):
        self.__hostaddr = hostaddr
        self.__hostport = hostport
        self.screens = {}
        self.broadcast_screen = Screen(0xffff)
        self._bind_screen(self.broadcast_screen)
        self.screens[self.broadcast_screen.index] = self.broadcast_screen
        self.tallies = {}
        self.__loop: asyncio.AbstractEventLoop|None = None
        self.running = False
        self.__connect_lock: asyncio.Lock|None = None
        self.__connected_evt: asyncio.Event|None = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """The :class:`asyncio.BaseEventLoop` associated with the instance"""
        loop = self.__loop
        if loop is None:
            loop = self.__loop = asyncio.get_running_loop()
        return loop

    @property
    def connected_evt(self) -> asyncio.Event:
        e = self.__connected_evt
        if e is None:
            e = self.__connected_evt = asyncio.Event()
        return e

    @property
    def _connect_lock(self) -> asyncio.Lock:
        l = self.__connect_lock
        if l is None:
            l = self.__connect_lock = asyncio.Lock()
        return l

    @property
    def hostaddr(self) -> str:
        """The local host address to bind the server to
        """
        return self.__hostaddr

    @property
    def hostport(self) -> int:
        """The port to listen on
        """
        return self.__hostport

    async def open(self):
        """Open the server
        """
        async with self._connect_lock:
            if self.running:
                return
            logger.debug('UmdReceiver.open()')
            self.running = True
            self.connected_evt.clear()
            self.transport, self.protocol = await self.loop.create_datagram_endpoint(
                lambda: UmdProtocol(self),
                local_addr=(self.hostaddr, self.hostport),
                reuse_port=True,
            )
            await self.connected_evt.wait()
            logger.info('UmdReceiver running')

    async def close(self):
        """Close the server
        """
        async with self._connect_lock:
            if not self.running:
                return
            logger.debug('UmdReceiver.close()')
            self.running = False
            self.transport.close()
            self.connected_evt.clear()
            logger.info('UmdReceiver closed')

    async def set_bind_address(self, hostaddr: str, hostport: int):
        """Set the :attr:`hostaddr` and :attr:`hostport` and restart the server
        """
        if hostaddr == self.hostaddr and hostport == self.hostport:
            return
        running = self.running
        if running:
            await self.close()
        self.__hostaddr = hostaddr
        self.__hostport = hostport
        if running:
            await self.open()

    async def set_hostaddr(self, hostaddr: str):
        """Set the :attr:`hostaddr` and restart the server
        """
        await self.set_bind_address(hostaddr, self.hostport)

    async def set_hostport(self, hostport: int):
        """Set the :attr:`hostport` and restart the server
        """
        await self.set_bind_address(self.hostaddr, hostport)

    def parse_incoming(self, data: bytes, addr: Tuple[str, int]):
        """Parse data received by the server
        """
        while True:
            message, remaining = Message.parse(data)
            if message.screen not in self.screens:
                screen = Screen(message.screen)
                self.screens[screen.index] = screen
                self._bind_screen(screen)
                self.emit('on_screen_added', screen)
                logger.debug(f'new screen: {screen.index}')
            else:
                screen = self.screens[message.screen]

            if message.is_broadcast:
                for screen in self.screens.values():
                    screen.update_from_message(message)
            else:
                screen.update_from_message(message)
            if not len(remaining):
                break

    def _bind_screen(self, screen: Screen):
        screen.bind(
            on_tally_added=self._on_screen_tally_added,
            on_tally_update=self._on_screen_tally_update,
            on_tally_control=self._on_screen_tally_control,
            on_control=self._on_screen_control,
        )

    def _on_screen_tally_added(self, tally: Tally, **kwargs):
        if tally.id not in self.tallies:
            self.tallies[tally.id] = tally
        self.emit('on_tally_added', tally, **kwargs)

    def _on_screen_tally_update(self, *args, **kwargs):
        self.emit('on_tally_updated', *args, **kwargs)

    def _on_screen_tally_control(self, *args, **kwargs):
        self.emit('on_tally_control', *args, **kwargs)

    def _on_screen_control(self, *args, **kwargs):
        self.emit('on_scontrol', *args, **kwargs)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    umd = UmdReceiver()

    loop.run_until_complete(umd.open())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(umd.close())
    finally:
        loop.close()



try:
    import loguru               # type: ignore[missing-import]
    from loguru import logger   # type: ignore[missing-import]
except ImportError:             # pragma: no cover
    loguru = None
    import logging
    logger = logging.getLogger(__name__)
import functools
import inspect

if loguru is not None:
    logger_catch = logger.catch # type: ignore[assignment]
else:
    def logger_catch(f):
        if inspect.iscoroutinefunction(f):
            @functools.wraps(f)
            async def async_wrapper(*args, **kwargs):
                try:
                    return await f(*args, **kwargs)
                except Exception as exc:
                    logger.exception(exc)
                    # raise
            return async_wrapper

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                f(*args, **kwargs)
            except Exception as exc:
                logger.error(f'Error in {f!r}', exc_info=True)
        return wrapper



from __future__ import annotations
import enum
from typing import Tuple, Iterator

__all__ = ('TallyColor', 'TallyType', 'TallyState', 'MessageType', 'TallyKey')

class TallyColor(enum.IntFlag):
    """Color enum for tally indicators

    Since this is an :class:`~enum.IntFlag`, its members can be combined using
    bitwise operators, making :attr:`AMBER` a combination of
    :attr:`RED` and :attr:`GREEN`

    This allows merging one color with another

    >>> from tslumd import TallyColor
    >>> TallyColor.RED
    <TallyColor.RED: 1>
    >>> TallyColor.GREEN
    <TallyColor.GREEN: 2>
    >>> TallyColor.AMBER
    <TallyColor.AMBER: 3>
    >>> TallyColor.RED | TallyColor.GREEN
    <TallyColor.AMBER: 3>


    .. versionchanged:: 0.0.4
        Bitwise operators
    """
    OFF = 0             #: Off
    RED = 1             #: Red
    GREEN = 2           #: Green
    AMBER = RED | GREEN #: Amber

    @staticmethod
    def from_str(s: str) -> TallyColor:
        """Return the member matching the given name (case-insensitive)

        >>> TallyColor.from_str('RED')
        <TallyColor.RED: 1>
        >>> TallyColor.from_str('green')
        <TallyColor.GREEN: 2>
        >>> TallyColor.from_str('Amber')
        <TallyColor.AMBER: 3>

        .. versionadded:: 0.0.5
        """
        return getattr(TallyColor, s.upper())

    def to_str(self) -> str:
        """The member name as a string

        >>> TallyColor.RED.to_str()
        'RED'
        >>> TallyColor.GREEN.to_str()
        'GREEN'
        >>> TallyColor.AMBER.to_str()
        'AMBER'
        >>> (TallyColor.RED | TallyColor.GREEN).to_str()
        'AMBER'

        .. versionadded:: 0.0.5
        """
        assert self.name is not None
        return self.name

    def __str__(self) -> str:
        return self.to_str()

    @classmethod
    def all(cls):
        """Iterate over all members

        .. versionadded:: 0.0.6
        """
        yield from cls.__members__.values()

    def __format__(self, format_spec: str) -> str:
        if format_spec == '':
            return str(self)
        return format(self.to_str(), format_spec)


class TallyType(enum.IntFlag):
    """Enum for the three tally display types in the UMD protocol

    Since this is an :class:`~enum.IntFlag`, its members can be combined using
    bitwise operators. The members can then be iterated over to retrieve the
    individual "concrete" values of :attr:`rh_tally`, :attr:`txt_tally`
    and :attr:`lh_tally`

    >>> from tslumd import TallyType
    >>> list(TallyType.rh_tally)
    [<TallyType.rh_tally: 1>]
    >>> list(TallyType.rh_tally | TallyType.txt_tally)
    [<TallyType.rh_tally: 1>, <TallyType.txt_tally: 2>]
    >>> list(TallyType.all_tally)
    [<TallyType.rh_tally: 1>, <TallyType.txt_tally: 2>, <TallyType.lh_tally: 4>]

    .. versionchanged:: 0.0.4
        Added support for bitwise operators and member iteration
    """
    no_tally = 0  #: No-op
    rh_tally = 1  #: :term:`Right-hand tally <rh_tally>`
    txt_tally = 2 #: :term:`Text tally <txt_tally>`
    lh_tally = 4  #: :term:`Left-hand tally <lh_tally>`
    all_tally = rh_tally | txt_tally | lh_tally
    """Combination of all tally types

    .. versionadded:: 0.0.4
    """

    @property
    def is_iterable(self) -> bool:
        """Returns ``True`` if this is a combination of multiple members

        (meaning it must be iterated over)

        .. versionadded:: 0.0.5
        """
        if self == TallyType.all_tally:
            return True
        if self.value == 0:
            return False
        mask = 1 << (self.bit_length() - 1)
        return self ^ mask != 0


    @classmethod
    def all(cls) -> Iterator[TallyType]:
        """Iterate over all members, excluding :attr:`no_tally` and :attr:`all_tally`

        .. versionadded:: 0.0.4
        """
        for ttype in cls:
            if ttype != TallyType.no_tally and ttype != TallyType.all_tally:
                yield ttype

    @staticmethod
    def from_str(s: str) -> TallyType:
        """Create an instance from a string of member name(s)

        The string can be a single member or multiple member names separated by
        a "|". For convenience, the names may be shortened by omitting the
        ``"_tally"`` portion from the end ("rh" == "rh_tally", etc)

        >>> TallyType.from_str('rh_tally')
        <TallyType.rh_tally: 1>
        >>> TallyType.from_str('rh|txt_tally')
        <TallyType.rh_tally|txt_tally: 3>
        >>> TallyType.from_str('rh|txt|lh')
        <TallyType.all_tally: 7>
        >>> TallyType.from_str('all')
        <TallyType.all_tally: 7>

        .. versionadded:: 0.0.5
        """
        if '|' in s:
            result = TallyType.no_tally
            for name in s.split('|'):
                result |= TallyType.from_str(name)
            return result
        s = s.lower()
        if not s.endswith('_tally'):
            s = f'{s}_tally'
        return getattr(TallyType, s)

    def to_str(self) -> str:
        """Create a string representation suitable for use in :meth:`from_str`

        >>> tt = TallyType.rh_tally
        >>> tt.to_str()
        'rh_tally'
        >>> tt |= TallyType.txt_tally
        >>> tt.to_str()
        'rh_tally|txt_tally'
        >>> tt |= TallyType.lh_tally
        >>> tt.to_str()
        'all_tally'

        .. versionadded:: 0.0.5
        """
        if self == TallyType.all_tally:
            assert self.name is not None
            return self.name
        if self.is_iterable:
            return '|'.join((str(obj) for obj in self))
        assert self.name is not None
        return self.name

    def __iter__(self) -> Iterator[TallyType]:
        for ttype in self.all():
            if ttype in self:
                yield ttype

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}.{self.to_str()}: {self.value}>'

    def __str__(self) -> str:
        return self.to_str()


class TallyState(enum.IntFlag):
    OFF = 0     #: Off
    PREVIEW = 1 #: Preview
    PROGRAM = 2 #: Program

class MessageType(enum.Enum):
    """Message type

    .. versionadded:: 0.0.2
    """
    _unset = 0
    display = 1 #: A message containing tally display information
    control = 2 #: A message containing control data

TallyKey = Tuple[int, int]
"""A tuple of (:attr:`screen_index <.Screen.index>`,
:attr:`tally_index <.Tally.index>`) to uniquely identify a single :class:`.Tally`
within its :class:`.Screen`
"""



from __future__ import annotations
try:
    from loguru import logger   # type: ignore[missing-import]
except ImportError:             # pragma: no cover
    import logging
    logger = logging.getLogger(__name__)
from typing import Union, Tuple, Iterator, cast, TYPE_CHECKING

from pydispatch import Dispatcher, Property

from tslumd import MessageType, TallyType, TallyColor, TallyKey
if TYPE_CHECKING:
    from .messages import Display, Message

StrOrTallyType = Union[str, TallyType]
StrOrTallyColor = Union[str, TallyColor]

__all__ = ('Tally', 'Screen')

class Tally(Dispatcher):
    """A single tally object

    Properties:
        rh_tally (TallyColor): State of the :term:`right-hand tally <rh_tally>` indicator
        txt_tally (TallyColor): State of the :term:`text tally <txt_tally>` indicator
        lh_tally (TallyColor): State of the :term:`left-hand tally <lh_tally>` indicator
        brightness (int): Tally indicator brightness from 0 to 3
        text (str): Text to display
        control (bytes): Any control data received for the tally indicator
        normalized_brightness (float): The :attr:`brightness` value normalized
            as a float from ``0.0`` to ``1.0``

    :Events:
        .. event:: on_update(instance: Tally, props_changed: set[str])

            Fired when any property changes

        .. event:: on_control(instance: Tally, data: bytes)

            Fired when control data is received for the tally indicator

    .. versionadded:: 0.0.2
        The :event:`on_control` event

    .. versionchanged:: 0.0.5
        Added container emulation
    """
    screen: Screen|None
    """The parent :class:`Screen` this tally belongs to

    .. versionadded:: 0.0.3
    """
    rh_tally = cast(TallyColor, Property(TallyColor.OFF))
    txt_tally = cast(TallyColor, Property(TallyColor.OFF))
    lh_tally = cast(TallyColor, Property(TallyColor.OFF))
    brightness = cast(int, Property(3))
    normalized_brightness = cast(float, Property(1.))
    text = cast(str, Property(''))
    control = cast(bytes, Property(b''))
    _events_ = ['on_update', 'on_control']
    _prop_attrs = ('rh_tally', 'txt_tally', 'lh_tally', 'brightness', 'text', 'control')
    def __init__(self, index_: int, **kwargs):
        self.screen = kwargs.get('screen')
        self.__index = index_
        if self.screen is not None:
            self.__id = (self.screen.index, self.__index)
        else:
            self.__id = None
        self._updating_props = False
        self.update(**kwargs)
        self.bind(**{prop:self._on_prop_changed for prop in self._prop_attrs})

    @property
    def index(self) -> int:
        """Index of the tally object from 0 to 65534 (``0xfffe``)
        """
        return self.__index

    @property
    def id(self) -> TallyKey:
        """A key to uniquely identify a :class:`Tally` / :class:`Screen`
        combination.

        Tuple of (:attr:`Screen.index`, :attr:`Tally.index`)

        Raises:
            ValueError: If the :attr:`Tally.screen` is ``None``

        .. versionadded:: 0.0.3
        """
        if self.__id is None:
            raise ValueError(f'Cannot create id for Tally without a screen ({self!r})')
        return self.__id

    @property
    def is_broadcast(self) -> bool:
        """``True`` if the tally is to be "broadcast", meaning sent to all
        :attr:`display indices<.messages.Display.index>`.

        (if the :attr:`index` is ``0xffff``)

        .. versionadded:: 0.0.2
        """
        return self.index == 0xffff

    @classmethod
    def broadcast(cls, **kwargs) -> Tally:
        """Create a :attr:`broadcast <is_broadcast>` tally

        (with :attr:`index` set to ``0xffff``)

        .. versionadded:: 0.0.2
        """
        return cls(0xffff, **kwargs)

    @classmethod
    def from_display(cls, display: Display, **kwargs) -> Tally:
        """Create an instance from the given :class:`~.messages.Display` object
        """
        attrs = set(cls._prop_attrs)
        if display.type.name == 'control':
            attrs.discard('text')
        else:
            attrs.discard('control')
        kw = kwargs.copy()
        kw.update({attr:getattr(display, attr) for attr in cls._prop_attrs})
        return cls(display.index, **kw)

    def set_color(self, tally_type: StrOrTallyType, color: StrOrTallyColor):
        """Set the color property (or properties) for the given TallyType

        Sets the :attr:`rh_tally`, :attr:`txt_tally` or :attr:`lh_tally`
        properties matching the :class:`~.common.TallyType` value(s).

        If the given tally_type is a combination of tally types, all of the
        matched attributes will be set to the given color.

        Arguments:
            tally_type (TallyType or str): The :class:`~.common.TallyType` member(s)
                to set. Multiple types can be specified using
                bitwise ``|`` operators.

                If the argument is a string, it should be formatted as shown in
                :meth:`.TallyType.from_str`
            color (TallyColor or str): The :class:`~.common.TallyColor` to set, or the
                name as a string


        >>> from tslumd import Tally, TallyType, TallyColor
        >>> tally = Tally(0)
        >>> tally.set_color(TallyType.rh_tally, TallyColor.RED)
        >>> tally.rh_tally
        <TallyColor.RED: 1>
        >>> tally.set_color('lh_tally', 'green')
        >>> tally.lh_tally
        <TallyColor.GREEN: 2>
        >>> tally.set_color('rh_tally|txt_tally', 'green')
        >>> tally.rh_tally
        <TallyColor.GREEN: 2>
        >>> tally.txt_tally
        <TallyColor.GREEN: 2>
        >>> tally.set_color('all', 'off')
        >>> tally.rh_tally
        <TallyColor.OFF: 0>
        >>> tally.txt_tally
        <TallyColor.OFF: 0>
        >>> tally.lh_tally
        <TallyColor.OFF: 0>

        .. versionadded:: 0.0.4

        .. versionchanged:: 0.0.5
            Allow string arguments and multiple tally_type members
        """
        self[tally_type] = color

    def get_color(self, tally_type: StrOrTallyType) -> TallyColor:
        """Get the color of the given tally_type

        If tally_type is a combination of tally types, the color returned will
        be a combination all of the matched color properties.

        Arguments:
            tally_type (TallyType or str): :class:`~.common.TallyType` member(s)
                to get the color values from.

                If the argument is a string, it should be formatted as shown in
                :meth:`.TallyType.from_str`


        >>> tally = Tally(0)
        >>> tally.get_color('rh_tally')
        <TallyColor.OFF: 0>
        >>> tally.set_color('rh_tally', 'red')
        >>> tally.get_color('rh_tally')
        <TallyColor.RED: 1>
        >>> tally.set_color('txt_tally', 'red')
        >>> tally.get_color('rh_tally|txt_tally')
        <TallyColor.RED: 1>
        >>> tally.get_color('all')
        <TallyColor.RED: 1>
        >>> tally.set_color('lh_tally', 'green')
        >>> tally.get_color('lh_tally')
        <TallyColor.GREEN: 2>
        >>> tally.get_color('all')
        <TallyColor.AMBER: 3>

        .. versionadded:: 0.0.5
        """
        return self[tally_type]

    def merge_color(self, tally_type: TallyType, color: TallyColor):
        """Merge the color property (or properties) for the given TallyType
        using the :meth:`set_color` method

        Combines the existing color value with the one provided using a bitwise
        ``|`` (or) operation

        Arguments:
            tally_type (TallyType): The :class:`~.common.TallyType` member(s)
                to merge. Multiple types can be specified using
                bitwise ``|`` operators.
            color (TallyColor): The :class:`~.common.TallyColor` to merge

        .. versionadded:: 0.0.4
        """
        for ttype in tally_type:
            cur_color = self[ttype]
            new_color = cur_color | color
            if new_color == cur_color:
                continue
            self[ttype] = new_color

    def merge(self, other: Tally, tally_type: TallyType = TallyType.all_tally):
        """Merge the color(s) from another Tally instance into this one using
        the :meth:`merge_color` method

        Arguments:
            other (Tally): The Tally instance to merge with
            tally_type (TallyType, optional): The :class:`~.common.TallyType`
                member(s) to merge. Multiple types can be specified using
                bitwise ``|`` operators.
                Default is :attr:`~.common.TallyType.all_tally` (all three types)

        .. versionadded:: 0.0.4
        """
        for ttype in tally_type:
            color = other[ttype]
            self.merge_color(ttype, color)

    def update(self, **kwargs) -> set[str]:
        """Update any known properties from the given keyword-arguments

        Returns:
            set: The property names, if any, that changed
        """
        log_updated = kwargs.pop('LOG_UPDATED', False)
        props_changed = set()
        self._updating_props = True
        for attr in self._prop_attrs:
            if attr not in kwargs:
                continue
            val = kwargs[attr]
            if attr == 'control' and val != b'':
                if self.control == val:
                    # logger.debug(f'resetting control, {val=}, {self.control=}')
                    self.control = b''
            if getattr(self, attr) == val:
                continue
            props_changed.add(attr)
            setattr(self, attr, val)
            if attr == 'brightness':
                val = cast(int, val)
                self.normalized_brightness = val / 3
            if log_updated:
                logger.debug(f'{self!r}.{attr} = {val!r}')
        self._updating_props = False
        if 'control' in props_changed and self.control != b'':
            self.emit('on_control', self, self.control)
        if len(props_changed):
            self.emit('on_update', self, props_changed)
        return props_changed

    def update_from_display(self, display: Display) -> set[str]:
        """Update this instance from the values of the given
        :class:`~.messages.Display` object

        Returns:
            set: The property names, if any, that changed
        """
        attrs = set(self._prop_attrs)
        is_control = display.type.name == 'control'
        if is_control:
            attrs.discard('text')
        else:
            attrs.discard('control')
        kw = {attr:getattr(display, attr) for attr in attrs}
        kw['LOG_UPDATED'] = True
        props_changed = self.update(**kw)
        return props_changed

    def to_dict(self) -> dict:
        """Serialize to a :class:`dict`
        """
        d = {attr:getattr(self, attr) for attr in self._prop_attrs}
        d['index'] = self.index
        if self.screen is None:
            d['id'] = None
        else:
            d['id'] = self.id
        return d

    # def to_display(self) -> 'tslumd.messages.Display':
    #     """Create a :class:`~.messages.Display` from this instance
    #     """
    #     kw = self.to_dict()
    #     return Display(**kw)

    def _on_prop_changed(self, instance, value, **kwargs):
        if self._updating_props:
            return
        prop = kwargs['property']
        if prop.name == 'control' and value != b'':
            self.emit('on_control', self, value)
        if prop.name == 'brightness':
            value = cast(int, value)
            self.normalized_brightness = value / 3
        self.emit('on_update', self, set([prop.name]))

    def __getitem__(self, key: StrOrTallyType) -> TallyColor:
        if not isinstance(key, TallyType):
            key = TallyType.from_str(key)
        if key.is_iterable:
            color = TallyColor.OFF
            for tt in key:
                assert tt.name is not None
                color |= getattr(self, tt.name)
            return color
        assert key.name is not None
        return getattr(self, key.name)

    def __setitem__(self, key: StrOrTallyType, value: StrOrTallyColor):
        if not isinstance(key, TallyType):
            key = TallyType.from_str(key)
        if not isinstance(value, TallyColor):
            value = TallyColor.from_str(value)
        if key.is_iterable:
            for tt in key:
                assert tt.name is not None
                setattr(self, tt.name, value)
        else:
            assert key.name is not None
            setattr(self, key.name, value)

    def __eq__(self, other):
        if not isinstance(other, Tally):
            return NotImplemented
        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        if not isinstance(other, Tally):
            return NotImplemented
        return self.to_dict() != other.to_dict()

    def __repr__(self):
        return f'<{self.__class__.__name__}: ({self})>'

    def __str__(self):
        if self.__id is None:
            return f'{self.index} - "{self.text}"'
        return f'{self.id} - "{self.text}"'

class Screen(Dispatcher):
    """A group of :class:`Tally` displays

    Properties:
        scontrol(bytes): Any control data received for the screen

    :Events:
        .. event:: on_tally_added(tally: Tally)

            Fired when a new :class:`Tally` instance is added to the screen

        .. event:: on_tally_update(tally: Tally, props_changed: set[str])

            Fired when any :class:`Tally` property changes. This is a
            retransmission of :event:`Tally.on_update`

        .. event:: on_tally_control(tally: Tally, data: bytes)

            Fired when control data is received for a :class:`Tally` object.
            This is a retransmission of :event:`Tally.on_control`

        .. event:: on_control(instance: Screen, data: bytes)

            Fired when control data is received for the :class:`Screen` itself

    .. versionadded:: 0.0.3

    """

    tallies: dict[int, Tally]
    """Mapping of :class:`Tally` objects within the screen using their
    :attr:`~Tally.index` as keys
    """

    scontrol = Property(b'')

    _events_ = [
        'on_tally_added', 'on_tally_update', 'on_tally_control', 'on_control',
    ]
    def __init__(self, index_: int):
        self.__index = index_
        self.tallies = {}
        self.bind(scontrol=self._on_scontrol_prop)

    @property
    def index(self) -> int:
        """The screen index from 0 to 65534 (``0xFFFE``)
        """
        return self.__index

    @property
    def is_broadcast(self) -> bool:
        """``True`` if the screen is to be "broadcast", meaning sent to all
        :attr:`screen indices<.messages.Message.screen>`.

        (if the :attr:`index` is ``0xffff``)
        """
        return self.index == 0xffff

    @classmethod
    def broadcast(cls, **kwargs) -> 'Screen':
        """Create a :attr:`broadcast <is_broadcast>` :class:`Screen`

        (with :attr:`index` set to ``0xffff``)
        """
        return cls(0xffff, **kwargs)

    def broadcast_tally(self, **kwargs) -> Tally:
        """Create a temporary :class:`Tally` using :meth:`Tally.broadcast`

        Arguments:
            **kwargs: Keyword arguments to pass to the :class:`Tally` constructor

        Note:
            The tally object is not stored in :attr:`tallies` and no event
            propagation (:event:`on_tally_added`, :event:`on_tally_update`,
            :event:`on_tally_control`) is handled by the :class:`Screen`.
        """
        return Tally.broadcast(screen=self, **kwargs)

    def add_tally(self, index_: int, **kwargs) -> Tally:
        """Create a :class:`Tally` object and add it to :attr:`tallies`

        Arguments:
            index_: The tally :attr:`~Tally.index`
            **kwargs: Keyword arguments passed to create the tally instance

        Raises:
            KeyError: If the given ``index_`` already exists
        """
        if index_ in self:
            raise KeyError(f'Tally exists for index {index_}')
        tally = Tally(index_, screen=self, **kwargs)
        self._add_tally_obj(tally)
        return tally

    def get_or_create_tally(self, index_: int) -> Tally:
        """If a :class:`Tally` object matching the given index exists, return
        it. Otherwise create one and add it to :attr:`tallies`

        This method is similar to :meth:`add_tally` and it can be used to avoid
        exception handling. It does not however take keyword arguments and
        is only intended for object creation.
        """
        if index_ in self:
            return self[index_]
        return self.add_tally(index_)

    def _add_tally_obj(self, tally: Tally):
        assert not tally.is_broadcast
        self.tallies[tally.index] = tally
        tally.bind(
            on_update=self._on_tally_updated,
            on_control=self._on_tally_control,
        )
        self.emit('on_tally_added', tally)

    def update_from_message(self, msg: Message):
        """Handle an incoming :class:`~.Message`
        """
        if msg.screen != self.index and not msg.broadcast:
            return
        if msg.type == MessageType.control:
            self.scontrol = msg.scontrol
        else:
            for dmsg in msg.displays:
                self.handle_dmsg(dmsg)

    def handle_dmsg(self, dmsg: Display):
        if dmsg.is_broadcast:
            for tally in self:
                tally.update_from_display(dmsg)
        else:
            if dmsg.index not in self:
                tally = Tally.from_display(dmsg, screen=self)
                self._add_tally_obj(tally)
                if dmsg.type == MessageType.control:
                    tally.emit('on_control', tally, tally.control)
            else:
                tally = self[dmsg.index]
                tally.update_from_display(dmsg)

    def _on_tally_updated(self, *args, **kwargs):
        self.emit('on_tally_update', *args, **kwargs)

    def _on_tally_control(self, *args, **kwargs):
        self.emit('on_tally_control', *args, **kwargs)

    def _on_scontrol_prop(self, instance: 'Screen', value: bytes, **kwargs):
        if not len(value):
            return
        self.emit('on_control', self, value)

    def __getitem__(self, key: int) -> Tally:
        return self.tallies[key]

    def __contains__(self, key: int) -> bool:
        return key in self.tallies

    def keys(self) -> Iterator[int]:
        yield from sorted((k for k in self.tallies.keys() if k != 0xffff))

    def values(self) -> Iterator[Tally]:
        for key in self.keys():
            yield self[key]

    def items(self) -> Iterator[Tuple[int, Tally]]:
        for key in self.keys():
            yield key, self[key]

    def __iter__(self) -> Iterator[Tally]:
        yield from self.values()

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'

    def __str__(self):
        return f'{self.index}'



import struct
import pytest

from tslumd import TallyColor, Message, Display, MessageType
from tslumd.messages import (
    Flags, ParseError, DmsgParseError,
    DmsgControlParseError, MessageParseError, MessageLengthError,
)

@pytest.fixture
def message_with_lots_of_displays():
    msgobj = Message()
    # Message header byte length: 6
    # Dmsg header byte length: 4
    # Text length: 2(length bytes) + 9(chars) = 11
    # Each Dmsg total: 4 + 11 = 15
    # Dmsg's: 4096 * 15 = 61440
    # 4096 Dmsg's with Message header: 4096 * 15 + 6 = 61446 bytes
    for i in range(4096):
        msgobj.displays.append(Display(index=i, text=f'Foo {i:05d}'))
    return msgobj


def test_uhs_message(uhs500_msg_bytes, uhs500_msg_parsed):
    parsed, remaining = Message.parse(uhs500_msg_bytes)
    assert not len(remaining)
    assert parsed == uhs500_msg_parsed


def test_messages():
    msgobj = Message(version=1, screen=5)
    rh_tallies = [getattr(TallyColor, attr) for attr in ['RED','OFF','GREEN','AMBER']]
    lh_tallies = [getattr(TallyColor, attr) for attr in ['GREEN','RED','OFF','RED']]
    txt_tallies = [getattr(TallyColor, attr) for attr in ['OFF','GREEN','AMBER','GREEN']]
    txts = ['foo', 'bar', 'baz', 'blah']
    indices = [4,3,7,1]
    for i in range(4):
        disp = Display(
            index=indices[i], rh_tally=rh_tallies[i], lh_tally=lh_tallies[i],
            txt_tally=txt_tallies[i], text=txts[i], brightness=i,
        )
        msgobj.displays.append(disp)

    parsed, remaining = Message.parse(msgobj.build_message())
    assert not len(remaining)

    for i in range(len(rh_tallies)):
        disp1, disp2 = msgobj.displays[i], parsed.displays[i]
        assert disp1.rh_tally == disp2.rh_tally == rh_tallies[i]
        assert disp1.lh_tally == disp2.lh_tally == lh_tallies[i]
        assert disp1.txt_tally == disp2.txt_tally == txt_tallies[i]
        assert disp1.text == disp2.text == txts[i]
        assert disp1.index == disp2.index == indices[i]
        assert disp1 == disp2

    for attr in ['version', 'flags', 'screen', 'scontrol']:
        assert getattr(msgobj, attr) == getattr(parsed, attr)

    assert msgobj == parsed


def test_packet_length(faker, message_with_lots_of_displays):
    msgobj = message_with_lots_of_displays

    # Make sure the 2048 byte limit is respected
    with pytest.raises(MessageLengthError):
        _ = msgobj.build_message()

    # Ensure that the limit can be bypassed
    msg_bytes = msgobj.build_message(ignore_packet_length=True)
    parsed, remaining = Message.parse(msg_bytes)
    assert parsed == msgobj

    # Iterate over individual message packets and make sure we get all displays
    all_parsed_displays = []
    for msg_bytes in msgobj.build_messages():
        assert len(msg_bytes) <= 2048
        parsed, remaining = Message.parse(msg_bytes)
        assert not len(remaining)
        all_parsed_displays.extend(parsed.displays)

    assert len(all_parsed_displays) == len(msgobj.displays)
    for disp, parsed_disp in zip(msgobj.displays, all_parsed_displays):
        assert disp.index == parsed_disp.index
        assert disp.text == parsed_disp.text

    # Create an SCONTROL that exceeds the limit
    msgobj = Message(scontrol=faker.binary(length=2048))
    with pytest.raises(MessageLengthError):
        it = msgobj.build_messages()
        _ = next(it)

    # Create a Dmsg control that exceeds the limit
    msgobj = Message(displays=[Display(index=0, control=faker.binary(length=2048))])
    with pytest.raises(MessageLengthError):
        it = msgobj.build_messages()
        _ = next(it)


def test_broadcast_message(faker):
    for i in range(1000):
        # Create Messages with random `screen` value in the non-broadcast range
        # and ensure the `is_broadcast` field is correct in both the
        # instance and its parsed version
        screen = faker.pyint(max_value=0xfffe)
        msgobj = Message(screen=screen)
        msgobj.displays.append(Display(index=i))
        assert not msgobj.is_broadcast

        parsed, remaining = Message.parse(msgobj.build_message())
        assert not msgobj.is_broadcast

    # Create broadcast Messages using both methods and check the `is_broadcast`
    # field on the instances and their parsed versions
    msgobj1 = Message(screen=0xffff)
    msgobj1.displays.append(Display(index=1))
    assert msgobj1.is_broadcast
    parsed1, remaining = Message.parse(msgobj1.build_message())
    assert parsed1.is_broadcast

    msgobj2 = Message.broadcast(displays=[Display(index=1)])
    assert msgobj2.is_broadcast
    parsed2, remaining = Message.parse(msgobj2.build_message())
    assert parsed2.is_broadcast

    assert msgobj1 == msgobj2 == parsed1 == parsed2


def test_broadcast_display(uhs500_msg_parsed, faker):

    disp_attrs = ('rh_tally', 'txt_tally', 'lh_tally', 'text', 'brightness')
    msgobj = Message()

    for uhs_disp in uhs500_msg_parsed.displays:
        assert not uhs_disp.is_broadcast

        # General kwargs excluding the `index`
        kw = {attr:getattr(uhs_disp, attr) for attr in disp_attrs}

        # Create random Displays within non-broadcast range and check the
        # `is_broadcast` field of the instance and its parsed version
        for _ in range(1000):
            ix = faker.pyint(max_value=0xfffe)
            disp = Display(index=ix, **kw)
            assert not disp.is_broadcast

            parsed, remaining = Display.from_dmsg(msgobj.flags, disp.to_dmsg(msgobj.flags))
            assert not parsed.is_broadcast
            assert parsed == disp

        # Create broadcast Displays using both methods and check the
        # `is_broadcast` field on the instances and their parsed versions
        bc_disp1 = Display.broadcast(**kw)
        bc_disp2 = Display(index=0xffff, **kw)
        assert bc_disp1.is_broadcast
        assert bc_disp2.is_broadcast

        parsed1, remaining = Display.from_dmsg(msgobj.flags, bc_disp1.to_dmsg(msgobj.flags))
        assert parsed1.is_broadcast

        parsed2, remaining = Display.from_dmsg(msgobj.flags, bc_disp2.to_dmsg(msgobj.flags))
        assert parsed2.is_broadcast

        assert bc_disp1 == bc_disp2 == parsed1 == parsed2

        # Add the broadcast Display to the Message at the top
        msgobj.displays.append(bc_disp1)

    # Check the `is_broadcast` field in the displays after Message building / parsing
    parsed, remaining = Message.parse(msgobj.build_message())
    for parsed_disp, bc_disp in zip(parsed.displays, msgobj.displays):
        assert parsed_disp.is_broadcast
        assert parsed_disp == bc_disp



def test_scontrol(faker):
    for _ in range(100):
        data_len = faker.pyint(min_value=1, max_value=1024)
        control_data = faker.binary(length=data_len)

        msgobj = Message(scontrol=control_data)
        assert msgobj.type == MessageType.control
        assert Flags.SCONTROL in msgobj.flags

        msg_bytes = msgobj.build_message()
        parsed, remaining = Message.parse(msg_bytes)
        assert not len(remaining)

        assert parsed.type == MessageType.control
        assert parsed.scontrol == control_data
        assert parsed == msgobj

        disp = Display(index=1)

        with pytest.raises(ValueError) as excinfo:
            disp_msg = Message(displays=[disp], scontrol=control_data)
        assert 'SCONTROL' in str(excinfo.value)


def test_dmsg_control(uhs500_msg_parsed, faker):
    tested_zero = False
    for _ in range(10):
        msgobj = Message(version=1, screen=5)
        for orig_disp in uhs500_msg_parsed.displays:
            if not tested_zero:
                data_len = 0
                tested_zero = True
            else:
                data_len = faker.pyint(min_value=0, max_value=1024)
            control_data = faker.binary(length=data_len)

            kw = orig_disp.to_dict()
            del kw['text']
            kw['control'] = control_data
            if not len(control_data):
                kw['type'] = MessageType.control
            disp = Display(**kw)

            assert disp.type == MessageType.control

            disp_bytes = disp.to_dmsg(msgobj.flags)
            parsed_disp, remaining = Display.from_dmsg(msgobj.flags, disp_bytes)

            assert not len(remaining)
            assert parsed_disp.control == control_data
            assert parsed_disp == disp

            msgobj.displays.append(disp)

        parsed = None
        for msg_bytes in msgobj.build_messages():
            _parsed, remaining = Message.parse(msg_bytes)
            assert not len(remaining)
            if parsed is None:
                parsed = _parsed
            else:
                parsed.displays.extend(_parsed.displays)
        assert parsed == msgobj

        with pytest.raises(ValueError) as excinfo:
            disp = Display(index=1, control=b'foo', text='foo')
        excstr = str(excinfo.value).lower()
        assert 'control' in excstr and 'text' in excstr

        with pytest.raises(ValueError) as excinfo:
            disp = Display(index=1, text='foo', type=MessageType.control)
        excstr = str(excinfo.value).lower()
        assert 'control' in excstr and 'text' in excstr


def test_invalid_message(uhs500_msg_bytes, faker):
    bad_bytes = faker.binary(length=5)
    with pytest.raises(MessageParseError) as excinfo:
        r = Message.parse(bad_bytes)
    assert 'header' in str(excinfo.value).lower()

    bad_bytes = bytearray(uhs500_msg_bytes)
    bad_byte_count = struct.pack('<H', len(uhs500_msg_bytes) + 10)
    bad_bytes[:2] = bad_byte_count
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(MessageParseError) as excinfo:
        r = Message.parse(bad_bytes)
    assert 'byte count' in str(excinfo.value).lower()


def test_invalid_dmsg(uhs500_msg_bytes, faker):

    # Clip the dmsg header fields
    bad_bytes = bytearray(uhs500_msg_bytes[:8])

    # Insert the correct value for `PBC` field so it gets past initial checks
    bad_byte_count = struct.pack('<H', len(bad_bytes) - 2)
    bad_bytes[:2] = bad_byte_count
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(DmsgParseError) as excinfo:
        r = Message.parse(bad_bytes)
    assert 'dmsg length' in str(excinfo.value).lower()

    # Clip the display text length field to the wrong size
    bad_bytes = bytearray(uhs500_msg_bytes[:10])

    # Insert the correct value for `PBC` field so it gets past initial checks
    bad_byte_count = struct.pack('<H', len(bad_bytes) - 2)
    bad_bytes[:2] = bad_byte_count
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(DmsgParseError) as excinfo:
        r = Message.parse(bad_bytes)
    assert 'text length' in str(excinfo.value).lower()

    # Insert an incorrect value for the text length field
    bad_bytes = bytearray(uhs500_msg_bytes)
    txt_len_bytes = struct.pack('<H', len(uhs500_msg_bytes) + 10)
    bad_bytes[10:12] = txt_len_bytes
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(DmsgParseError) as excinfo:
        r = Message.parse(bad_bytes)
    assert 'invalid text bytes' in str(excinfo.value).lower()


def test_invalid_dmsg_control(uhs500_msg_bytes, faker):
    msg = Message()
    disp = Display(index=1, control=b'foo\x00')
    msg.displays.append(disp)
    msg_bytes = msg.build_message()

    # Clip the length field to the wrong size
    bad_bytes = bytearray(msg_bytes)
    bad_bytes = bad_bytes[:-5]
    bad_byte_count = struct.pack('<H', len(bad_bytes) - 2)
    bad_bytes[:2] = bad_byte_count
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(DmsgControlParseError):
        r = Message.parse(bad_bytes)

    # Clip the control bytes to the wrong length
    bad_bytes = bytearray(msg_bytes)
    bad_bytes = bad_bytes[:-2]
    bad_byte_count = struct.pack('<H', len(bad_bytes) - 2)
    bad_bytes[:2] = bad_byte_count
    bad_bytes = bytes(bad_bytes)
    with pytest.raises(DmsgControlParseError):
        r = Message.parse(bad_bytes)



from __future__ import annotations
import asyncio
import dataclasses
from dataclasses import dataclass, field
import enum
import struct
from typing import Tuple, Iterator, Any, cast

from tslumd import MessageType, TallyColor, Tally

__all__ = (
    'Display', 'Message', 'ParseError', 'MessageParseError',
    'DmsgParseError', 'DmsgControlParseError', 'MessageLengthError',
)


class ParseError(Exception):
    """Raised on errors during message parsing

    .. versionadded:: 0.0.2
    """
    msg: str #: Error message
    value: bytes #: The relevant message bytes containing the error
    def __init__(self, msg: str, value: bytes):
        self.msg = msg
        self.value = value
    def __str__(self):
        return f'{self.msg}: "{self.value!r}"'

class MessageParseError(ParseError):
    """Raised on errors while parsing :class:`Message` objects

    .. versionadded:: 0.0.2
    """
    pass

class DmsgParseError(ParseError):
    """Raised on errors while parsing :class:`Display` objects

    .. versionadded:: 0.0.2
    """
    pass

class DmsgControlParseError(ParseError):
    """Raised on errors when parsing :attr:`Display.control` data

    .. versionadded:: 0.0.2
    """
    pass

class MessageLengthError(ValueError):
    """Raised when message length is larger than 2048 bytes

    .. versionadded:: 0.0.4
    """


class Flags(enum.IntFlag):
    """Message flags
    """
    NO_FLAGS = 0 #: No flags set
    UTF16 = 1
    """Indicates text formatted as ``UTF-16LE`` if set, otherwise ``UTF-8``"""

    SCONTROL = 2
    """Indicates the message contains ``SCONTROL`` data if set, otherwise ``DMESG``
    """


@dataclass
class Display:
    """A single tally "display"
    """
    index: int #: The display index from 0 to 65534 (``0xFFFE``)
    rh_tally: TallyColor = TallyColor.OFF #: Right hand tally indicator
    txt_tally: TallyColor = TallyColor.OFF #: Text tally indicator
    lh_tally: TallyColor = TallyColor.OFF #: Left hand tally indicator
    brightness: int = 3 #: Display brightness (from 0 to 3)
    text: str = '' #: Text to display
    control: bytes = b''
    """Control data (if :attr:`type` is :attr:`~.MessageType.control`)

    .. versionadded:: 0.0.2
    """

    type: MessageType = MessageType.display
    """The message type. One of :attr:`~.MessageType.display` or
    :attr:`~.MessageType.control`.

    * For :attr:`~.MessageType.display` (the default), the message contains
      :attr:`text` information and the :attr:`control` field must be empty.
    * For :attr:`~.MessageType.control`, the message contains :attr:`control`
      data and the :attr:`text` field must be empty

    .. versionadded:: 0.0.2
    """

    is_broadcast: bool = field(init=False)
    """``True`` if the display is to a "broadcast", meaning sent to all display
    indices.

    (if the :attr:`index` is ``0xffff``)

    .. versionadded:: 0.0.2
    """

    def __post_init__(self):
        self.is_broadcast = self.index == 0xffff
        if len(self.control):
            self.type = MessageType.control
        if self.type == MessageType.control and len(self.text):
            raise ValueError('Control message cannot contain text')

    @classmethod
    def broadcast(cls, **kwargs) -> Display:
        """Create a :attr:`broadcast <is_broadcast>` display

        (with :attr:`index` set to ``0xffff``)

        .. versionadded:: 0.0.2
        """
        kwargs = kwargs.copy()
        kwargs['index'] = 0xffff
        return cls(**kwargs)

    @classmethod
    def from_dmsg(cls, flags: Flags, dmsg: bytes) -> Tuple[Display, bytes]:
        """Construct an instance from a ``DMSG`` portion of received message.

        Any remaining message data after the relevant ``DMSG`` is returned along
        with the instance.
        """
        if len(dmsg) < 4:
            raise DmsgParseError('Invalid dmsg length', dmsg)
        hdr = struct.unpack('<2H', dmsg[:4])
        hdr = cast(Tuple[int, int], hdr)
        dmsg = dmsg[4:]
        ctrl = hdr[1]
        kw: dict[str, Any] = dict(
            index=hdr[0],
            rh_tally=TallyColor(ctrl & 0b11),
            txt_tally=TallyColor(ctrl >> 2 & 0b11),
            lh_tally=TallyColor(ctrl >> 4 & 0b11),
            brightness=ctrl >> 6 & 0b11,
        )
        is_control_data = ctrl & 0x8000 == 0x8000
        if is_control_data:
            ctrl, dmsg = cls._unpack_control_data(dmsg)
            kw['control'] = ctrl
            kw['type'] = MessageType.control
        else:
            if len(dmsg) < 2:
                raise DmsgParseError('Invalid text length field', dmsg)
            txt_byte_len = struct.unpack('<H', dmsg[:2])[0]
            dmsg = dmsg[2:]
            txt_bytes = dmsg[:txt_byte_len]
            dmsg = dmsg[txt_byte_len:]
            if len(txt_bytes) != txt_byte_len:
                raise DmsgParseError(
                    f'Invalid text bytes. Expected {txt_byte_len}',
                    txt_bytes,
                )
            if Flags.UTF16 in flags:
                txt = txt_bytes.decode('UTF-16le')
            else:
                if b'\0' in txt_bytes:
                    txt_bytes = txt_bytes.split(b'\0')[0]
                txt = txt_bytes.decode('UTF-8')
            kw['text'] = txt
        return cls(**kw), dmsg

    @staticmethod
    def _unpack_control_data(data: bytes) -> Tuple[bytes, bytes]:
        """Unpack control data (if control bit 15 is set)

        Arguments:
            data: The portion of the ``dmsg`` at the start of the
                "Control Data" field

        Returns:
            bytes: remaining
                The remaining message data after the control data field

        Note:
            This is undefined as of UMDv5.0 and its implementation is
            the author's "best guess" based off of other areas of the protocol

        .. versionadded:: 0.0.2

        :meta public:
        """
        if len(data) < 2:
            raise DmsgControlParseError('Unknown control data format', data)
        length = struct.unpack('<H', data[:2])[0]
        data = data[2:]
        if len(data) < length:
            raise DmsgControlParseError('Unknown control data format', data)
        return data[:length], data[length:]

    @staticmethod
    def _pack_control_data(data: bytes) -> bytes:
        """Pack control data (if control bit 15 is set)

        Arguments:
            data: The control data to pack

        Returns:
            bytes: packed
                The packed control data

        Note:
            This is undefined as of UMDv5.0 and its implementation is
            the author's "best guess" based off of other areas of the protocol

        .. versionadded:: 0.0.2

        :meta public:
        """
        length = len(data)
        return struct.pack(f'<H{length}s', length, data)

    def to_dmsg(self, flags: Flags) -> bytes:
        """Build ``dmsg`` bytes to be included in a message
        (called from :meth:`Message.build_message`)
        """
        ctrl = self.rh_tally & 0b11
        ctrl += (self.txt_tally & 0b11) << 2
        ctrl += (self.lh_tally & 0b11) << 4
        ctrl += (self.brightness & 0b11) << 6
        if self.type == MessageType.control:
            ctrl |= 0x8000
            data = bytearray(struct.pack('<2H', self.index, ctrl))
            data.extend(self._pack_control_data(self.control))
        else:
            if Flags.UTF16 in flags:
                txt_bytes = bytes(self.text, 'UTF-16le')
            else:
                txt_bytes = bytes(self.text, 'UTF-8')
            txt_byte_len = len(txt_bytes)
            data = bytearray(struct.pack('<3H', self.index, ctrl, txt_byte_len))
            data.extend(txt_bytes)
        return data

    def to_dict(self) -> dict:
        d = dataclasses.asdict(self)
        del d['is_broadcast']
        return d

    @classmethod
    def from_tally(cls, tally: Tally, msg_type: MessageType = MessageType.display) -> Display:
        """Create a :class:`Display` from the given :class:`~.Tally`

        .. versionadded:: 0.0.2
            The msg_type argument
        """
        kw = tally.to_dict()
        del kw['id']
        if msg_type == MessageType.control:
            del kw['text']
        elif msg_type == MessageType.display:
            del kw['control']
        kw['type'] = msg_type
        return cls(**kw)

    def __eq__(self, other):
        if not isinstance(other, (Display, Tally)):
            return NotImplemented
        self_dict = self.to_dict()
        oth_dict = other.to_dict()
        if isinstance(other, Display):
            return self_dict == oth_dict
        else:
            del oth_dict['id']

        del self_dict['type']
        if self.type == MessageType.control:
            del self_dict['text']
            del oth_dict['text']
        else:
            del self_dict['control']
            del oth_dict['control']

        return self_dict == oth_dict

    def __ne__(self, other):
        if not isinstance(other, (Display, Tally)):
            return NotImplemented
        return not self.__eq__(other)

@dataclass
class Message:
    """A single UMDv5 message packet
    """
    version: int = 0 #: Protocol minor version
    flags: Flags = Flags.NO_FLAGS #: The message :class:`Flags` field
    screen: int = 0 #: Screen index from 0 to 65534 (``0xFFFE``)
    displays: list[Display] = field(default_factory=list)
    """A list of :class:`Display` instances"""

    scontrol: bytes = b''
    """SCONTROL data (if :attr:`type` is :attr:`~.MessageType.control`)"""

    type: MessageType = MessageType.display
    """The message type. One of :attr:`~.MessageType.display` or
    :attr:`~.MessageType.control`.

    * For :attr:`~.MessageType.display` (the default), the contents of
      :attr:`displays` are used and the :attr:`scontrol` field must be empty.
    * For :attr:`~.MessageType.control`, the :attr:`scontrol` field is used and
      :attr:`displays` must be empty.

    .. versionadded:: 0.0.2
    """

    is_broadcast: bool = field(init=False)
    """``True`` if the message is to be "broadcast" to all screens.

    (if :attr:`screen` is ``0xffff``)

    .. versionadded:: 0.0.2
    """

    def __post_init__(self):
        self.is_broadcast = self.screen == 0xffff
        if not isinstance(self.flags, Flags):
            self.flags = Flags(self.flags)

        if len(self.scontrol) and len(self.displays):
            raise ValueError('SCONTROL message cannot contain displays')

        if len(self.scontrol):
            self.type = MessageType.control

        if self.type == MessageType.control:
            self.flags |= Flags.SCONTROL
        elif self.type == MessageType._unset:
            if Flags.SCONTROL in self.flags:
                self.type = MessageType.control
            else:
                self.type = MessageType.display

    @classmethod
    def broadcast(cls, **kwargs) -> Message:
        """Create a :attr:`broadcast <is_broadcast>` message

        (with :attr:`screen` set to ``0xffff``)

        .. versionadded:: 0.0.2
        """
        kwargs = kwargs.copy()
        kwargs['screen'] = 0xffff
        return cls(**kwargs)

    @classmethod
    def parse(cls, msg: bytes) -> Tuple[Message, bytes]:
        """Parse incoming message data to create a :class:`Message` instance.

        Any remaining message data after parsing is returned along with the instance.
        """
        if len(msg) < 6:
            raise MessageParseError('Invalid header length', msg)
        data = struct.unpack('<HBBH', msg[:6])
        byte_count, version, flags, screen = data
        kw = dict(
            version=version,
            flags=Flags(flags),
            screen=screen,
            type=MessageType._unset,
        )
        msg = msg[2:]
        if len(msg) < byte_count:
            raise MessageParseError(
                f'Invalid byte count. Expected {byte_count}, got {len(msg)}',
                msg,
            )
        remaining = msg[byte_count:]
        msg = msg[4:byte_count]
        obj = cls(**kw)
        if obj.type == MessageType.control:
            obj.scontrol = msg
            return obj, remaining
        while len(msg):
            disp, msg = Display.from_dmsg(obj.flags, msg)
            obj.displays.append(disp)
        return obj, remaining

    def build_message(self, ignore_packet_length: bool = False) -> bytes:
        """Build a message packet from data in this instance

        Arguments:
            ignore_packet_length (bool, optional): If ``False``, the message limit
                of 2048 bytes is respected, and if exceeded, an exception is raised.
                Otherwise, the limit is ignored. (default is False)

        Raises:
            MessageLengthError: If the message packet is larger than 2048 bytes
                (and ``ignore_packet_length`` is False)

        Note:
            This method is retained for backwards compatability. To properly
            handle the message limit, use :meth:`build_messages`

        .. versionchanged:: 0.0.4

            * The ``ignore_packet_length`` parameter was added
            * Message length is limited to 2048 bytes
        """
        it = self.build_messages(ignore_packet_length=ignore_packet_length)
        data = next(it)
        try:
            next_data = next(it)
        except StopIteration:
            pass
        else:
            if not ignore_packet_length:
                raise MessageLengthError()
        return data

    def build_messages(self, ignore_packet_length: bool = False) -> Iterator[bytes]:
        """Build message packet(s) from data in this instance as an iterator

        The specified maximum packet length of 2048 is respected and if
        necessary, the data will be split into separate messages.

        This method will always function as a :term:`generator`, regardless of
        the number of message packets produced.

        .. versionadded:: 0.0.4
        """
        msg_len_exceeded = False
        next_disp_index = None
        if self.type == MessageType.control:
            payload = bytearray(self.scontrol)
            byte_count = len(payload)
            if byte_count + 6 > 2048:
                raise MessageLengthError()
        else:
            byte_count = 0
            payload = bytearray()
            for disp_index, display in enumerate(self.displays):
                disp_payload = display.to_dmsg(self.flags)
                disp_len = len(disp_payload)
                if not ignore_packet_length:
                    if byte_count + disp_len + 6 >= 2048:
                        if disp_index == 0:
                            raise MessageLengthError()
                        msg_len_exceeded = True
                        next_disp_index = disp_index
                        break
                byte_count += disp_len
                payload.extend(disp_payload)
        fmt = f'<HBBH{byte_count}B'
        pbc = struct.calcsize(fmt) - 2
        data = bytearray(struct.pack('<HBBH', pbc, self.version, self.flags, self.screen))
        data.extend(payload)
        yield bytes(data)

        if msg_len_exceeded:
            displays = self.displays[next_disp_index:]
            attrs = ('version', 'flags', 'screen', 'scontrol', 'type')
            kw = {attr:getattr(self, attr) for attr in attrs}
            kw['displays'] = displays
            sub_msg = Message(**kw)
            yield from sub_msg.build_messages()
